"""
Bootstrax: XENONnT online processing manager
=============================================
How to use
----------------
    <activate conda environment>
    bootstrax --production

    or with the preferred settings:
    bootstrax --production --infer_mode --delete_live --undying

----------------

For more info, see the documentation:
https://straxen.readthedocs.io/en/latest/bootstrax.html
"""

__version__ = "2.0.1"

import os
import os.path as osp
import typing as ty
import argparse
import typing
from datetime import datetime, timedelta, timezone
import logging
import multiprocessing
import npshmex
import signal
import socket
import shutil
import time
import traceback
import numpy as np
import pymongo
from psutil import pid_exists, disk_usage, Process
import pytz
import strax
import straxen
import threading
import pandas as pd
import daqnt
import fnmatch
from glob import glob
from straxen import daq_core
from straxen.daq_core import now


# Patch for targeted (uncompressed) chunk size
straxen.Peaklets.chunk_target_size_mb = strax.DEFAULT_CHUNK_SIZE_MB
straxen.nVETOHitlets.chunk_target_size_mb = strax.DEFAULT_CHUNK_SIZE_MB

parser = argparse.ArgumentParser(description="XENONnT online processing manager")
parser.add_argument(
    "--debug", action="store_true", help="Start strax processes with debug logging."
)
parser.add_argument(
    "--profile",
    type=str,
    default="false",
    help="Run strax in profiling mode. Specify target filename as argument.",
)
parser.add_argument(
    "--cores", type=int, default=8, help="Maximum number of workers to use in a strax process."
)
parser.add_argument(
    "--targets",
    nargs="*",
    default="event_info events_nv events_mv online_peak_monitor veto_proximity".split(),
    help="Strax data type name(s) that should be produced with live processing",
)
parser.add_argument(
    "--post_process",
    nargs="*",
    default=["veto_intervals"],
    help=(
        "Target(s) for other sub-detectors. If not produced automatically "
        "when processing tpc data, st.make the requested data later."
    ),
)
parser.add_argument(
    "--fix_target",
    action="store_true",
    help="Don't allow bootstrax to switch to a different target for special runs",
)
parser.add_argument(
    "--fix_resources",
    action="store_true",
    help="Don't let bootstrax change number of cores/max_messages because of failures",
)
parser.add_argument(
    "--infer_mode",
    action="store_true",
    help="Determine best number max-messages and cores for each run "
    "automatically. Overrides --cores and --max_messages",
)
parser.add_argument(
    "--delete_live",
    action="store_true",
    help="Delete live_data after successful processing of the run.",
)
parser.add_argument(
    "--production",
    action="store_true",
    help=(
        "Run bootstrax in production mode. Assuming test mode otherwise to "
        "prevent interactions with the runs-database"
    ),
)
parser.add_argument(
    "--ignore_checks",
    action="store_true",
    help="Do not use! This disables checks on e.g. the timestamps! Should only "
    "be used if some run is very valuable but some checks are failing.",
)
parser.add_argument("--max_messages", type=int, default=10, help="number of max mailbox messages")


actions = parser.add_mutually_exclusive_group()
actions.add_argument(
    "--process", type=int, metavar="NUMBER", help="Process a single run, regardless of its status."
)
actions.add_argument(
    "--fail",
    nargs="+",
    metavar=("NUMBER", "REASON"),
    help="Fail run number, optionally with reason",
)
actions.add_argument(
    "--abandon",
    nargs="+",
    metavar=("NUMBER", "REASON"),
    help="Abandon run number, optionally with reason",
)
actions.add_argument("--undying", action="store_true", help="Except any error and ignore it")
args = parser.parse_args()

##
# Configuration
##

# The folder that can be used for testing bootstrax (i.e. non production
# mode). It will be written to:
test_data_folder = (
    "/data/test_processed/" if os.path.exists("/data/test_processed/") else "./bootstrax/"
)

# Timeouts in seconds
timeouts = {
    # Waiting between escalating SIGTERM -> SIGKILL -> crashing bootstrax
    # when trying to kill another process (usually child strax)
    "signal_escalate": 3,
    # Minimum waiting time to retry a failed run
    # Escalates exponentially on repeated failures: 1x, 5x, 25x, 125x, 125x, 125x, ...
    # Some jitter is applied: actual delays will randomly be 0.5 - 1.5x as long
    "retry_run": 60,
    # Maximum time [s] for strax to complete a processing
    # if exceeded, strax will be killed by bootstrax
    "max_processing_time": 72 * 3600,
    # Max processing time [s] for bootstrax to finish processing after run ended
    "max_processing_time_after_completion": 13 * 3600,
    # Every this many [s] we should have at least one new file written to disk
    "max_no_write_time": 30 * 60,
    # Sleep between checking whether a strax process is alive
    "check_on_strax": 10,
    # Maximum time a run is 'busy' without a further update from
    # its responsible bootstrax. Bootstrax normally updates every
    # check_on_strax seconds, so make sure this is substantially
    # larger than check_on_strax.
    "max_busy_time": 120,
    # Maximum time a run is in the 'considering' state
    # if exceeded, will be labeled as an untracked failure
    "max_considering_time": 60,
    # Minimum time to wait between database cleanup operations
    "cleanup_spacing": 60,
    # Sleep time when there is nothing to do
    "idle_nap": 10,
    # If we don't hear from a bootstrax on another host for this long,
    # remove its entry from the bootstrax status collection
    # Must be much longer than idle_nap and check_on_strax!
    "bootstrax_presumed_dead": 300,
    # Ebs3-5 normally do all the processing. However if all are busy
    # for a longer period of time, the ebs0-2 can also help with
    # processing.
    "eb3-5_max_busy_time": 5 * 60,
    # Bootstrax writes it's state to the daq-database. To have a backlog we store this
    # state using a TTL collection. To prevent too many entries in this backlog, only
    # create new entries if the previous entry is at least this old (in seconds).
    "min_status_interval": 60,
    # Maximum time we can take to can infer the datarate (s).
    "max_data_rate_infer_time": 60,
    # Minimum time we have to be in the run before we can infer the datarate (s).
    "min_data_rate_infer_time": 7.5,
    # Bootstrax can abandon runs for this time based on a tag, after that it
    # should not be on the DAQ any way or can be manually failed using the
    # --abandon option. We can abandon a run only for this many seconds:
    "abandoning_allowed": 3600 * 24 * 1,
}

# The disk that the eb is writing to may fill up at some point. The data should
# be written to datamanager at some point. This may clean up data on the disk,
# hence, we can check if there is sufficient diskspace and if not, wait a while.
# Below are the max number of times and number of seconds bootstrax will wait.
wait_diskspace_max_space_percent = 95
min_disk_space_tb = 1.0  # Terabytes
wait_diskspace_n_max = 60 * 24 * 7  # times
wait_diskspace_dt = 10  # seconds
if timeouts["bootstrax_presumed_dead"] < wait_diskspace_dt:
    raise ValueError("wait_diskspace_dt too large")

# Fields in the run docs that bootstrax uses. Pay attention to the tailing spaces!
bootstrax_projection = (
    "name start end number bootstrax status mode detectors rate "
    "data.host "
    "data.type "
    "data.location "
    "daq_config.processing_threads "
    "daq_config.compressor "
    "daq_config.strax_fragment_payload_bytes "
    "daq_config.strax_chunk_length "
    "daq_config.strax_chunk_overlap".split()
)

# Filename for temporary storage of the exception
# This is used to communicate the exception from the strax child process
# to the bootstrax main process
exception_tempfile = "last_bootstrax_exception.txt"

# The name of the thread that is opened to delete live_data
delete_thread_name = "DeleteThread"

# boostrax state for 'dead' or old entries in the bs_coll
dead_state = "dead_bootstrax"

# The maximum time difference (s) allowed between the timestamps in the data and the
# duration of the run (from the runs metadeta). Fail if the difference is larger than:
max_timetamp_diff = 5

# The maximum number of retries for processing a run. After this many times of retrying
# to process a run, the DAQ-group has to either manually fix this run or manually fail it.
max_n_retry = 10

# Check that there are runs that are waiting to be processed. If there are few,
# an eventbuilder should not process new data.
max_queue_new_runs = 3

# Remove any targets or post processing targets after the run failed
# this many times. If high level data is hitting some edge case, we
# might want to be able to keep the intermediate level data.
# NB: fnmatch so event* applies e.g. to event_basics, peak* to e.g. peaklets
# See https://xe1t-wiki.lngs.infn.it/doku.php?id=cfuselli:bootstrax_fail_targets
remove_target_after_fails = {
    "event*": 2,
    "corrected_areas": 2,
    "energy_estimates": 2,
    "distinct_channels": 2,
    "*pos*": 3,
    "individual_peak*": 3,
    "online_monitor_*v": 3,
    "peak_*": 4,
    "online_peak*": 4,
    "*hit*": 4,
    "veto_*": 4,
    "*pulse*": 4,
    "led_*": 4,
    "ext_timings_nv": 4,
    "merged_s2s*": 5,
    "peak*": 6,
    "lone_*": 6,
    "detector_time_*": 6,
    "*": 7,
}

##
# Initialize globals (e.g. rundb connection)
##
hostname = socket.getfqdn()

versions = straxen.print_versions(
    modules="strax straxen utilix daqnt numpy tensorflow numba".split(),
    include_git=True,
    return_string=True,
)

log_name = "bootstrax_" + hostname + ("" if args.production else "_TESTING")
log = daqnt.get_daq_logger(
    log_name,
    log_name,
    level=logging.DEBUG,
    opening_message=f"I am processing with these software versions: {versions}",
)

# Set the output folder
output_folder = daq_core.pre_folder if args.production else test_data_folder

if not args.production:
    # This means we are in some test mode
    wait_diskspace_max_space_percent = 80
    if not os.path.exists(output_folder):
        log.warning(f"Creating {output_folder}")
        os.mkdir(output_folder)
    log.warning(
        "\n---------------"
        "\nBe aware, bootstrax not running in production mode. Specify with --production."
        f"\nWriting new data to {output_folder}. Not saving this location in the RunDB."
        "\nNot writing to the runs-database."
        "\n---------------"
    )
    time.sleep(5)
else:
    if not args.delete_live:
        log.warning(
            "Production mode is designed to run with '--delete_live'\nplease restart bootstrax"
        )
    if not args.infer_mode:
        log.warning(
            "Better performance is expected in production mode with "
            "'--infer_mode'\nplease restart bootstrax"
        )

if os.access(output_folder, os.W_OK) is not True:
    message = f"No writing access to {output_folder}"
    log.warning(message)
    raise IOError(message)


def new_context(
    cores=args.cores,
    max_messages=args.max_messages,
    timeout=500,
):
    """Create strax context that can access the runs db."""
    # We use exactly the same logic of straxen to access the runs DB;
    # this avoids duplication, and ensures strax can access the runs DB if we can

    context = straxen.contexts.xenonnt_online(
        output_folder=output_folder,
        we_are_the_daq=True,
        allow_multiprocess=cores > 1,
        allow_shm=cores > 1,
        allow_lazy=False,
        max_messages=max_messages,
        timeout=timeout,
        _rucio_path=None,
    )
    if not args.production:
        # Keep the rundb but set it to readonly and local only, delete
        # all other storage frontends except fo the test folder.
        context.storage = [context.storage[0], strax.DataDirectory(output_folder)]
        context.storage[0].readonly = True
        context.storage[0].local_only = True
    return context


st = new_context()

databases = daq_core.DataBases(production=args.production)
run_db = databases.run_db
daq_db = databases.daq_db
run_coll = databases.run_coll
log_coll = databases.log_coll
ag_stat_coll = databases.ag_stat_coll
bs_coll = databases.bs_coll
run_db.command("ping")

# Ping the databases to ensure the mongo connections are working
if not args.undying:
    run_db.command("ping")
    daq_db.command("ping")


def run():
    if args.cores == -1:
        # Use all of the available cores on this machine
        args.cores = multiprocessing.cpu_count()
        log.info(f"Set cores to n_tot, using {args.cores} cores")

    if args.fail:
        args.fail += [""]  # Provide empty reason if none specified
        manual_fail(number=int(args.fail[0]), reason=args.fail[1])

    elif args.abandon:
        number = int(args.abandon[0])
        if len(args.abandon) > 1:
            manual_fail(number=number, reason=args.abandon[1])
        abandon(number=number)

    elif args.process:
        t_start = now()
        number = args.process

        # Check whether the run is already processed
        bootstrax_state = (
            run_coll.find_one({"number": number}, projection={"bootstrax": True})
            .get("bootstrax", {})
            .get("state", "no-state")
        )
        if args.production and bootstrax_state in ["done", "busy", "considering"]:
            message = f"It looks like run {number} is already processed."
            log_warning(message, priority="fatal")
            raise ValueError(message)

        rd = consider_run({"number": number})
        if rd is None:
            message = f"Trying to process single run but no run numbered {number} exists"
            log_warning(message, priority="fatal")
            raise ValueError(message)

        set_state("busy")
        process_run(rd)
        log.info(
            f"bootstrax ({hostname}) finished run {number} in"
            f" {(now() - t_start).total_seconds()} seconds"
        )
        wait_on_delete_thread()

    else:
        # Start processing
        loop()


##
# Main loop
##


def loop():
    """Infinite loop looking for runs to process."""
    # Ensure we're the only bootstrax on this host
    any_other_running = list(bs_coll.find({"host": hostname, "pid": {"$ne": os.getpid()}}))
    for x in any_other_running:
        if pid_exists(x["pid"]) and x["pid"]:
            log.warning(f'Bootstrax already running with PID {x["pid"]}, trying to kill it.')
            kill_process(x["pid"])

    # # Register ourselves
    set_state("starting")
    t_start = now()

    next_cleanup_time = now()
    # keep track of the ith run that we have seen when we are not in production mode
    new_runs_seen, failed_runs_seen = 0, 1
    while True:
        log.info(f"bootstrax running for {(now() - t_start).total_seconds()} seconds")
        # Check resources are still OK, otherwise crash / reboot program
        sufficient_diskspace()
        log.info("Looking for work")
        set_state("busy")
        if eb_can_process():
            # Process new runs
            rd = consider_run({"bootstrax.state": None}, test_counter=new_runs_seen)
            if rd is not None:
                new_runs_seen += 1
                process_run(rd)
                continue
        else:
            # We are on an old eb with not so much to do, perhaps one of
            # the veto systems needs processing?
            rd = consider_run(
                {"detectors": {"$ne": "tpc"}, "bootstrax.state": None}, test_counter=new_runs_seen
            )
            if rd is not None:
                new_runs_seen += 1
                process_run(rd)
                continue

        # There is either no new run or we are an old eventbuilder.
        # Scan DB for runs with unusual problems
        if now() > next_cleanup_time:
            cleanup_db()
            next_cleanup_time = now(plus=timeouts["cleanup_spacing"])

        # Any failed runs to retry?
        # Only try one run, we want to be back for new runs quickly
        rd = consider_run(
            {
                "bootstrax.state": "failed",
                "bootstrax.n_failures": {"$lt": max_n_retry},
                "bootstrax.next_retry": {"$lt": now()},
            },
            test_counter=failed_runs_seen,
        )

        if rd is not None:
            failed_runs_seen += 1
            process_run(rd)
            continue
        # Nothing to do, let's do some cleanup
        if not args.production:
            log.info(
                "We have gone through the rundDB in a readonly mode there are no "
                f"runs left. We looked at {new_runs_seen} new runs and "
                f"{failed_runs_seen} previously failed runs."
            )
            break
        log.info("No work to do, waiting for new runs or retry timers")
        set_state("idle")
        time.sleep(timeouts["idle_nap"])


##
# General helpers
##


def kill_process(pid):
    """Kill process pid."""
    log.warning(f"Kill PID:{pid}")
    if not pid_exists(pid):
        log.warning(f"No PID:{pid}")
        return

    parent = Process(pid)
    for child in parent.children(recursive=True):
        child.kill()
    parent.kill()

    # Just make it extra dead
    os.kill(pid, signal.SIGKILL)

    if pid_exists(pid):
        message = f"Could not kill process {pid}?!"
        log_warning(message, priority="fatal")


def _remove_veto_from_t(
    targets: ty.Union[str, list, tuple],
    remove: ty.Union[str, list, tuple] = ("_mv", "_nv"),
    _flip: bool = False,
) -> ty.Union[str, list, tuple, None]:
    """Remove veto(s) from targets."""
    start = strax.to_str_tuple(targets)
    remove = strax.to_str_tuple(remove)
    if targets is None:
        return None
    for r in remove:
        targets = keep_target(targets, {f"*{r}": 0}, 1)
    if _flip:
        targets = [t for i, t in enumerate(start) if not np.in1d(start, targets)[i]]
    return strax.to_str_tuple(targets)


def _keep_veto_from_t(
    targets: ty.Union[str, list, tuple],
    keep: ty.Union[str, list, tuple] = "_nv",
) -> ty.Union[str, list, tuple, None]:
    """Remove non-veto(s) targets."""
    targets = _remove_veto_from_t(targets, remove=keep, _flip=True)  # type: ignore
    if not len(targets):
        targets = strax.to_str_tuple("raw_records")
    return targets


def keep_target(targets, compare_with, n_fails):
    kept_targets = []
    delete_after = -1  # just to make logging never fail below
    for target_name in strax.to_str_tuple(targets):
        for delete_target, delete_after in compare_with.items():
            failed_too_much = n_fails > delete_after
            name_matches = fnmatch.fnmatch(target_name, delete_target)
            if failed_too_much and name_matches:
                log.warning(f"remove {target_name} ({n_fails}>{delete_after})")
                break
        else:
            log.debug(f"keep {target_name} ({n_fails}!>{delete_after})")
            kept_targets.append(target_name)
    if not len(kept_targets):
        kept_targets = ["raw_records"]
    return kept_targets


def infer_target(rd: dict) -> dict:
    """Check if the target should be overridden based on the mode of the DAQ for this run.

    :param rd: rundoc
    :return: dict with the targets and the targets for post processing

    """
    targets = args.targets.copy()
    post_process = args.post_process.copy()

    if args.fix_target:
        return {
            "targets": strax.to_str_tuple(targets),
            "post_processing": strax.to_str_tuple(post_process),
        }

    n_fails = rd["bootstrax"].get("n_failures", 0)

    if n_fails:
        log.debug(f"Deleting targets")
        targets = keep_target(targets, remove_target_after_fails, n_fails)
        post_process = keep_target(post_process, remove_target_after_fails, n_fails)

    log.debug(f"{targets} and {post_process} remaining")

    # Special modes override target for these
    led_modes = ["pmtgain"]
    diagnostic_modes = [
        "exttrig",
        "noise",
        "mv_diffuserballs",
        "mv_fibres",
        "mv_darkrate",
    ]
    ap_modes = ["pmtap"]
    nv_ref_mon = [
        "nVeto_LASER_calibration",
    ]
    mode = str(rd.get("mode"))
    detectors = list(rd.get("detectors"))  # type: ignore

    log.debug(f"mode is {mode}, changing target if needed")
    if np.any([m in mode for m in led_modes]):
        log.debug("led-mode")
        targets = "led_calibration"
        post_process = "raw_records"
    elif np.any([m in mode for m in ap_modes]):
        log.debug("afterpulse mode")
        targets = "afterpulses"
        post_process = "raw_records"
    elif np.any([m in mode for m in nv_ref_mon]):
        log.debug("NV reflecitvity and diffuser ball mode")
        targets = "ref_mon_nv"
        post_process = "raw_records"
    elif np.any([m in mode for m in diagnostic_modes]):
        log.debug("diagnostic-mode")
        targets = "raw_records"
        post_process = "raw_records"
    elif "kr83m" in mode and (len(targets) or len(post_process)):
        # Override the first (highest level) plugin for Kr runs (could
        # also use source field, outcome is the same)
        if "event_info" in targets or "event_info" in post_process:
            targets = list(targets) + ["event_info_double"]
    elif "ambe" in mode:
        # rates are very high, to ensure smooth operation let's just do this
        # based on calibrations of Apr 2023 this is the only safe working solution
        log.debug("ambe-mode")

        # get the mode from the daq_db
        # this is a new thing from Nov 2023
        # it overwrites the mode from the rundb
        bootstrax_config_coll = daq_db["bootstrax_config"]
        bootstrax_config = bootstrax_config_coll.find_one({"name": "bootstrax_config"})

        this_eb_ambe_mode = bootstrax_config["ambe_modes"].get(hostname[:3], "default")
        log.debug(f"Ambe mode for {hostname} is {this_eb_ambe_mode}")

        if this_eb_ambe_mode != "default":
            log.debug(f"Overwriting targets and post processing for {hostname} from daq_db")
            targets = bootstrax_config["modes_definitions"][this_eb_ambe_mode]["targets"]
            post_process = bootstrax_config["modes_definitions"][this_eb_ambe_mode]["post_process"]

    targets = strax.to_str_tuple(targets)
    post_process = strax.to_str_tuple(post_process)

    if "tpc" not in detectors:
        keep = []
        if "neutron_veto" in detectors:
            keep += ["_nv"]
        if "muon_veto" in detectors:
            keep += ["_mv"]
        targets = _keep_veto_from_t(targets, keep=keep)
        post_process = _keep_veto_from_t(post_process, keep=keep)
    else:
        for det, remove in (("neutron_veto", "_nv"), ("muon_veto", "_mv")):
            if det not in detectors:
                # Remove the _veto if this detector is not in the detector list
                targets = _remove_veto_from_t(targets, remove=remove)
                post_process = _remove_veto_from_t(post_process, remove=remove)
        if len(detectors) > 1:
            log.info(
                f'{rd["number"]:06} running in linked mode ({detectors}), '
                f"processing up to {targets} and postprocessing "
                f"to {post_process}"
            )

    if targets is None or not len(targets):
        targets = "raw_records"
    if post_process is None or not len(post_process):
        post_process = "raw_records"

    targets = strax.to_str_tuple(targets)
    post_process = strax.to_str_tuple(post_process)
    log.info(f"Inferring modes done, writing {targets} and {post_process}")
    for check in (targets, post_process):
        if not len(set(check)) == len(check):
            log_warning(f"Duplicates in (post) targets {check}", priority="fatal")
            raise ValueError(f"Duplicates in (post) targets {check}")

    return {"targets": targets, "post_processing": post_process}


def set_state(state, update_fields=None):
    """Inform the bootstrax collection we're in a different state.

    if state is None, leave state unchanged, just update heartbeat time

    """
    # Find the last message of this host
    previous_entry = bs_coll.find_one({"host": hostname}, sort=[("_id", pymongo.DESCENDING)])
    if state is None:
        state = "None" if previous_entry is None else previous_entry.get("state")

    bootstrax_state = dict(
        host=hostname,
        pid=os.getpid(),
        time=now(),
        state=state,
        targets=args.targets,
        max_cores=args.cores,
        max_messages=args.max_messages,
        undying=args.undying,
        production_mode=args.production,
    )
    if update_fields:
        update_fields = strax.storage.mongo.remove_np(update_fields)
        bootstrax_state.update(update_fields)

    need_new_doc = (previous_entry is None) or (
        (now() - previous_entry["time"].replace(tzinfo=pytz.utc)).seconds
        > timeouts["min_status_interval"]
    )
    if need_new_doc:
        bs_coll.insert_one(bootstrax_state)
    else:
        bs_coll.update_one({"_id": previous_entry["_id"]}, {"$set": bootstrax_state})


def send_heartbeat(update_fields=None):
    """Inform the bootstrax collection we're still here Use during long-running tasks where state
    doesn't change."""
    # Same as set_state, just don't change state
    set_state(None, update_fields=update_fields)


def log_warning(message, priority="warning", run_id=None):
    getattr(log, priority.lower())(message)
    databases.log_warning(
        message,
        priority=priority,
        run_id=run_id,
        production=args.production,
        user=f"bootstrax_{hostname}",
    )


def eb_can_process():
    """The new ebs (eb3-5) should be sufficient to process all data. In exceptional circumstances
    eb3-5 cannot keep up. Only let eb0-2 also process data in such cases.

    Before eb0-2 are also used for processing two criteria have to be fulfilled:
        - There should be runs waiting to be processed
        - Eb3-5 should be busy processing for a substantial time.
    :return: bool if this host should process a run

    """

    # eb3-5 always process.
    if hostname in ["eb3.xenon.local", "eb4.xenon.local", "eb5.xenon.local"]:
        return True

    # In test mode we can always process
    if not args.production:
        return True
    elif "eb2" in hostname:
        log_warning("Why is eb2 alive?!", priority="error")
        return False

    # Count number of runs untouched by bootstrax.
    n_untouched_runs = run_coll.count_documents({"bootstrax.state": None})

    # Check that eb3-5 are all busy for at least some time.
    n_ebs_running = 0
    n_ebs_busy = 0
    for eb_i in range(3, 6):
        # Should count if eb3-5 are registered as running (as one might be offline).
        bootstrax_on_host = bs_coll.find_one(
            {
                "host": f"eb{eb_i}.xenon.local",
                "time": {"$gt": now(-timeouts["bootstrax_presumed_dead"])},
            },
            sort=[("time", pymongo.DESCENDING)],
        )

        if bootstrax_on_host:
            n_ebs_running += 1
            running_eb = run_coll.find_one(
                {
                    "bootstrax.state": "busy",
                    "bootstrax.host": f"eb{eb_i}.xenon.local",
                    "bootstrax.started_processing": {"$lt": now(-timeouts["eb3-5_max_busy_time"])},
                }
            )
            if running_eb:
                n_ebs_busy += 1
                log.debug(f"eb{eb_i} is busy")
    log.info(f"running: {n_ebs_running}\tbusy: {n_ebs_busy}\tqueue: {n_untouched_runs}")
    if not n_ebs_running:
        return True
    if n_untouched_runs > max_queue_new_runs:
        return True
    return False


def infer_mode(rd):
    """Infer a safe operating mode of running bootstrax based on the uncompressed redax rate.

    Estimating save parameters for running
    bootstrax from:
    https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:daq:eb_speed_tests_2021update
    :return: dictionary of how many cores, max_messages and compressor
    should be used based on an estimated data rate.

    """
    # Get data rate from dispatcher
    data_rate = 0
    if "rate" in rd:
        # When a run finishes, the rundb is updated with the rates by the
        # dispatcher. Especially useful if the aggregate status does not
        # have the run anymore.
        data_rate = float(sum([detector["max"] for detector in rd["rate"].values()]))
    try:
        started_looking = time.time()
        time_to_wait = (
            timeouts["min_data_rate_infer_time"]
            - (now() - rd["start"].replace(tzinfo=timezone.utc)).total_seconds()
        )
        if time_to_wait > 0:
            log.debug(f"Waiting {time_to_wait:.1f} s to infer datarate")
            time.sleep(time_to_wait)
        while data_rate == 0:
            # For runs that are still running, we should be able to get the
            # info from the aggregate status collection.
            docs = ag_stat_coll.aggregate(
                [
                    {"$match": {"number": rd["number"]}},
                    {"$group": {"_id": "$detector", "rate": {"$max": "$rate"}}},
                ]
            )
            data_rate = float(sum([d["rate"] for d in docs]))
            if data_rate > 0:
                break
            elif time.time() - started_looking > timeouts["max_data_rate_infer_time"]:
                raise RuntimeError(f'Could not infer_mode for {rd["number"]}')
            log.debug(
                f'No rate inferred for {rd["number"]} after {time.time() - started_looking:.1f} s.'
            )
            time.sleep(2)

    except Exception as e:
        log_warning(
            f"infer_mode ran into {e}. Cannot infer datarate, using default mode.",
            run_id=f'{rd["number"]:06}',
            priority="warning",
        )
        data_rate = 0

    # Find out if eb is new (eb3-eb5):
    is_new_eb = int(hostname[2]) >= 3  # ebX.xenon.local
    log.info(f"Data rate: {data_rate:.1f} MB/s. New_eb: {is_new_eb}")
    benchmark = {
        "mbs": [0, 70, 90, 110, 150, 220, 290, 360, 390, 420, 500, 550],
        "cores_old": [39, 35, 35, 30, 30, 20, 12, 12, 10, 10, 10, 8],
        "cores": [24, 24, 24, 24, 18, 15, 15, 15, 15, 15, 15, 10],
        "max_messages_old": [20, 20, 15, 15, 10, 10, 10, 10, 10, 10, 8, 6],
        "max_messages": [60, 60, 35, 30, 25, 25, 25, 25, 20, 15, 12, 12],
        "timeout": [1200, 1200, None, None, None, None, None, None, None, None, None, 2400],
    }
    if data_rate and args.infer_mode:

        # Temporary solution
        # It is a patch to try to process some ambe data without failing for memory
        # If we are doing ambe -> consider the maximum rate
        # so that we have 10 cores and 12 messages for new ebs
        # and we have     8 cores and 6 messages for old ebs
        # added by Carlo on 19 April 2023
        mode = str(rd.get("mode"))
        if "ambe" in mode:
            data_rate = 550

        df = pd.DataFrame(benchmark)
        if data_rate not in benchmark["mbs"]:
            df.loc[len(df.index)] = [data_rate, None, None, None, None, None]
        df.set_index("mbs", inplace=True)
        df.sort_values("mbs", inplace=True)
        df.interpolate(method="index", inplace=True)
        result = {k: int(v) for k, v in df.loc[data_rate].items()}
        if not is_new_eb:
            for k in ("cores", "max_messages"):
                result[k] = result[k + "_old"]
        del df, benchmark, result["cores_old"], result["max_messages_old"]
    else:
        result = dict(cores=args.cores, max_messages=args.max_messages, timeout=1000)

    n_fails = rd["bootstrax"].get("n_failures", 0)
    if args.fix_resources:
        # If we are in a fix resource mode, we should not change the resources
        # based on the number of failures.
        n_fails = 0
        log.debug(f"Fixing resources, ignoring {n_fails} previous failures")

    if n_fails:
        # Exponentially lower resources & increase timeout
        result = dict(
            cores=np.clip(result["cores"] / (1.1**n_fails), 4, 40).astype(int),
            max_messages=np.clip(result["max_messages"] / (1.1**n_fails), 4, 100).astype(int),
            timeout=np.clip(result["timeout"] * (1.1**n_fails), 500, 3600).astype(int),
        )
        log_warning(
            f'Repeated failures on {rd["number"]}@{hostname}. Lowering to {result}',
            priority="info",
            run_id=f'{rd["number"]:06}',
        )
    else:
        result = {k: int(v) for k, v in result.items()}
    result["records_compressor"] = infer_records_compressor(rd, data_rate, n_fails)
    log.info(f'Inferred mode for {rd["number"]}\t{result}')
    return result


def infer_records_compressor(rd, datarate, n_fails):
    """
    Get a compressor for the (raw)records. This takes two things in consideration:
    1. Do we store the data fast enough (high write speed)
    2. Does the data fit into the buffer

    Used compressors:
        bz2: slow but very good compression -> use for low datarate
        zstd: fast & decent compression, max chunk size of ??? GB
        lz4: fast & not no chunk size limit, use if all ese fails
    """
    if n_fails or datarate is None:
        # Cannot infer datarate or failed before, go for fast & safe
        return "lz4" if n_fails > 1 else "zstd"

    chunk_length = rd["daq_config"]["strax_chunk_overlap"] + rd["daq_config"]["strax_chunk_length"]
    chunk_size_mb = datarate * chunk_length
    if datarate < 65:
        # Low data rate, we can do very large compression
        return "zstd"
    if chunk_size_mb > 1800:
        # Extremely large chunks, let's use LZ4 because we know that it
        # can handle this.
        return "lz4"
    # High datarate and reasonable chunk size.
    return "zstd"


##
# Host interactions
##


def sufficient_diskspace():
    """Check if there is sufficient space available on the local disk to write to."""
    for i in range(wait_diskspace_n_max):
        du = disk_usage(output_folder)
        disk_pct = du.percent
        disk_free = du.free / (1024**4)
        if disk_pct < wait_diskspace_max_space_percent and disk_free > min_disk_space_tb:
            log.debug(f"Check disk space: {disk_pct:.1f}% full")
            # Sufficient space to write to, let's continue
            return
        if i == 0:
            # Log it once to the database, the first time. Otherwise, just log it locally
            log_warning(
                f"Insufficient free disk space ({disk_pct:.1f}% full) on {hostname}. "
                f"Waiting {i}/{wait_diskspace_n_max}",
                priority="warning",
            )
        else:
            log.warning(f"Insufficient free disk space ({disk_pct:.1f}% full)")
        time.sleep(wait_diskspace_dt)
        send_heartbeat(dict(state="disk full"))
    set_state(dead_state)
    message = f"No disk space to write to. Kill bootstrax on {hostname}"
    log_warning(message, priority="fatal")
    raise RuntimeError(message)


def delete_live_data(rd, live_data_path):
    """Open thread to delete the live_data."""
    if args.production and os.path.exists(live_data_path) and args.delete_live:
        delete_thread = threading.Thread(
            name=delete_thread_name, target=_delete_data, args=(rd, live_data_path, "live")
        )
        log.info(f"Starting thread to delete {live_data_path} at {now()}")
        # We rather not stop deleting the live_data if something else
        # fails. Set the thread to daemon.
        delete_thread.setDaemon(True)
        delete_thread.start()
        log.info(
            f"DeleteThread {live_data_path} should be running in parallel, "
            f"continue MainThread now: {now()}"
        )


def _delete_data(rd, path, data_type):
    """After completing the processing and updating the RunDB, remove the live_data."""

    if data_type == "live" and not args.delete_live and args.production:
        message = "Unsafe operation. Trying to delete live data!"
        log_warning(message, priority="fatal")
        raise ValueError(message)
    log.debug(f"Deleting data at {path}")
    if os.path.exists(path):
        shutil.rmtree(path)
    log.info(f"deleting {path} finished")
    # Remove the data location from the rundoc and append it to the 'deleted_data' entries
    if not os.path.exists(path):
        log.info("changing data field in rundoc")
        for ddoc in rd["data"]:
            if ddoc["type"] == data_type:
                break
        for k in ddoc.copy().keys():
            if k in ["location", "meta", "protocol"]:
                ddoc.pop(k)

        ddoc.update({"at": now(), "by": hostname})
        log.debug(f"update with {ddoc}")
        run_coll.update_one(
            {"_id": rd["_id"]},
            {
                "$addToSet": {"deleted_data": ddoc},
                "$pull": {"data": {"type": data_type, "host": {"$in": ["daq", hostname]}}},
            },
        )
    else:
        message = f"Something went wrong we wanted to delete {path}!"
        log_warning(message, priority="fatal")
        raise ValueError(message)


def wait_on_delete_thread():
    """Check that the thread with the delete_thread_name is finished before continuing."""
    threads = threading.enumerate()
    for thread in threads:
        if thread.name == delete_thread_name:
            done = False
            while not done:
                if thread.is_alive():
                    log.info(f'{thread.name} still running take a {timeouts["idle_nap"]} s nap')
                    time.sleep(timeouts["idle_nap"])
                    done = True
    log.info(f"Checked that {delete_thread_name} finished")


def clear_shm():
    """Manually delete files in /dev/shm/ created by npshmex on starup."""
    shm_dir = "/dev/shm/"
    shm_files = [f for f in os.listdir(shm_dir) if "npshmex" in f]

    if not len(shm_files):
        return
    log.info(f"Clearing {len(shm_files)} files")
    for f in shm_files:
        os.remove(shm_dir + f)


##
# Run DB interaction
##


def ping_dbs():
    while True:
        try:
            run_db.command("ping")
            daq_db.command("ping")
            break
        except Exception as ping_error:
            log.warning(
                f"Failed to connect to Mongo. Ran into {ping_error}. Sleep for a minute.",
                priority="warning",
            )
            time.sleep(60)


def get_run(*, mongo_id=None, number=None, full_doc=False):
    """Find and return run doc matching mongo_id or number The bootstrax state is left unchanged.

    :param full_doc: If true (default is False), return the full run doc rather than just fields
        used by bootstrax.

    """
    if number is not None:
        query = {"number": number}
    elif mongo_id is not None:
        query = {"_id": mongo_id}
    else:
        # This means you are not running a normal bootstrax (no reason to report to rundb)
        raise ValueError("Please give mongo_id or number")

    return run_coll.find_one(query, projection=None if full_doc else bootstrax_projection)


def set_run_state(rd, state, return_new_doc=True, **kwargs):
    """Set state of run doc rd to state
    return_new_doc: if True (default), returns new document.
        if False, instead returns the original (un-updated) doc.

    Any additional kwargs will be added to the bootstrax field.
    """
    if not args.production:
        return run_coll.find_one({"_id": rd["_id"]})

    bd = rd["bootstrax"]
    bd.update({"state": state, "host": hostname, "time": now(), **kwargs})

    if state == "failed":
        bd["n_failures"] = bd.get("n_failures", 0) + 1

    return run_coll.find_one_and_update(
        {"_id": rd["_id"]},
        {"$set": {"bootstrax": bd}},
        return_document=return_new_doc,
        projection=bootstrax_projection,
    )


def check_data_written(rd):
    """Checks that the data as written in the runs-database is actually available on this machine.

    :param rd: rundoc
    :return: type bool, False if not all paths exist or if there are no files on this host.

    """
    files_written = 0

    # Fetch the rd again -> to see status of 'data' field
    new_rd = get_run(mongo_id=rd["_id"], full_doc=True)
    log.debug(new_rd["data"])
    for ddoc in new_rd["data"]:
        ddoc_loc = ddoc.get("location", "NO LOCATION")
        ddoc_host = ddoc.get("host", "NO HOST")
        log.debug(f"Checking {ddoc_loc} on {ddoc_host} (current {hostname})")
        if ddoc_host == hostname:
            log.debug(f"Counting files {ddoc_loc} on {ddoc_host}")
            if os.path.exists(ddoc_loc):
                log.debug(f"{ddoc_loc} written")
                files_written += 1
            else:
                log.info(f"No data at {ddoc_loc}")
                return False
    log.info(f"{files_written} files are saved")
    return files_written > 0


def all_files_saved(rd, wait_max=600, wait_per_cycle=10):
    """Check that all files are written. It might be that the savers are still in the process of
    renaming from folder_temp to folder. Hence allow some wait time to allow the savers to finish.

    :param rd: rundoc
    :param wait_max: max seconds to wait for data to save
    :param wait_per_cycle: wait this many seconds if the data is not yet there

    """
    start = time.time()
    while not check_data_written(rd):
        log.debug(f'{rd["number"]} not all saved')
        if time.time() - start > wait_max:
            log_warning(f'Not all files saved for {rd["number"]}@{hostname}?!', priority="warning")
            return False
        send_heartbeat()
        time.sleep(wait_per_cycle)
    return True


def get_end_time(rd) -> typing.Optional[datetime]:
    return run_coll.find_one({"_id": rd["_id"]}, projection={"end": 1}).get("end", None)


def set_status_finished(rd):
    """Set the status to ready to upload for datamanager and admix."""
    # Check mongo connection
    ping_dbs()

    if not args.production:
        # Don't update the status if we are not in production mode
        return

    # First check that all the data is available (that e.g. no _temp files
    # are being renamed). This line should be over-redundant as we already
    # check earlier.
    all_files_saved(rd)

    # Only update the status if it does not exist or if it needs to be uploaded
    ready_for_restrax = {"status": "eb_finished_pre"}
    if rd.get("status") in [None, "needs_upload"]:
        run_coll.update_one({"_id": rd["_id"]}, {"$set": ready_for_restrax})
    elif rd.get("status") == ready_for_restrax.get("status"):
        # This is strange, bootstrax already finished this run before
        log_warning(
            "WARNING: bootstax has already marked this run as ready for restrax. Doing nothing.",
            priority="warning",
            run_id=f'{rd["number"]:06}',
        )
    else:
        # Do not override this field for runs already uploaded in admix
        message = (
            f'Trying to set set the status {rd.get("status")} to '
            f"{ready_for_restrax}! One should not override this field."
        )
        log_warning(message, priority="fatal")
        raise ValueError(message)


def abandon(*, mongo_id=None, number=None):
    """Mark a run as abandoned."""
    set_run_state(get_run(mongo_id=mongo_id, number=number), "abandoned")


def consider_run(query, return_new_doc=True, test_counter=0):
    """Return one run doc matching query, and simultaneously set its bootstrax state to
    'considering'."""
    # We must first do an atomic find-and-update to set the run's state
    # to "considering", to ensure the run doesn't get picked up by a
    # bootstrax on another host.
    if args.production:
        rd = run_coll.find_one_and_update(
            query,
            {"$set": {"bootstrax.state": "considering"}},
            projection=bootstrax_projection,
            return_document=True,
            sort=[("start", pymongo.DESCENDING)],
        )
        # Next, we can update the bootstrax entry properly with set_run_state
        # (adding hostname, time, etc.)
        if rd is None:
            return None
        return set_run_state(rd, "considering", return_new_doc=return_new_doc)
    else:
        # Don't change the runs-database for test modes
        try:
            rds = run_coll.find(
                query, projection=bootstrax_projection, sort=[("start", pymongo.DESCENDING)]
            )
            return rds[test_counter]
        except IndexError:
            return None


def fail_run(rd, reason, error_traceback=""):
    """Mark the run represented by run doc rd as failed with reason."""
    if "number" not in rd:
        long_run_id = f"run <no run number!!?>:{rd['_id']}"
    else:
        long_run_id = f"run {rd['number']}"

    # No bootstrax info is present when manually failing a run with args.fail
    if "bootstrax" not in rd.keys():
        rd["bootstrax"] = {}
        rd["bootstrax"]["n_failures"] = 0

    if (rd["bootstrax"].get("n_failures", 0) > 0) or (
        "perhaps it crashed on this run or is still stuck" in reason
    ):
        fail_name = "Repeated failure"
        failure_message_level = "info"
    else:
        fail_name = "New failure"
        failure_message_level = "warning"

    # Cleanup any data associated with the run
    # TODO: This should become optional, or just not happen at all,
    # after we're done testing (however, then we need some other
    # pruning mechanism like AJAX!)
    clean_run(mongo_id=rd["_id"])

    # Report to run db
    # It's best to do this after everything is done;
    # as it changes the run state back away from 'considering', so another
    # bootstrax could conceivably pick it up again.
    set_run_state(
        rd,
        "failed",
        reason=f"{hostname}:\n{reason}\n{error_traceback}",
        next_retry=(
            now(
                plus=(
                    timeouts["retry_run"]
                    * np.random.uniform(0.5, 1.5)
                    # Exponential backoff with jitter
                    * 5 ** min(rd["bootstrax"].get("n_failures", 0), 3)
                )
            )
        ),
    )

    # Report to DAQ log and screen. Let's not also add the entire traceback
    log_warning(
        f"{fail_name} on {long_run_id}: {reason}",
        priority=failure_message_level,
        run_id=f'{rd["number"]:06}',
    )


def manual_fail(*, mongo_id=None, number=None, reason=""):
    """Manually mark a run as failed based on mongo_id or run number."""
    rd = get_run(mongo_id=mongo_id, number=number)
    fail_run(rd, "Manually set failed state. " + reason)


##
# Processing
##


def run_strax(
    run_id,
    input_dir,
    targets,
    readout_threads,
    compressor,
    run_start_time,
    samples_per_record,
    cores,
    max_messages,
    timeout,
    daq_chunk_duration,
    daq_overlap_chunk_duration,
    post_processing,
    records_compressor,
    debug=False,
):
    # Check mongo connection
    ping_dbs()
    # Clear the swap memory used by npshmmex
    npshmex.shm_clear()
    # double check by forcefully clearing shm
    clear_shm()

    if debug:
        logging.basicConfig(force=True)
        logging.getLogger().setLevel(logging.DEBUG)
    try:
        log.info(f"Starting strax to make {run_id} with input dir {input_dir}")

        if targets == strax.to_str_tuple("led_calibration"):
            # TODO: still true?
            # timeout *= 5
            pass

        # Create multiple targets
        st = new_context(
            cores=cores,
            max_messages=max_messages,
            timeout=timeout,
        )

        for t in ("raw_records", "records", "records_nv", "hitlets_nv"):
            # Set the (raw)records processor to the inferred one
            st._plugin_class_registry[t].compressor = records_compressor

        # Make a function for running strax, call the function to process the run
        # This way, it can also be run inside a wrapper to profile strax
        def st_make():
            """Run strax."""
            strax_config = dict(
                daq_input_dir=input_dir,
                daq_compressor=compressor,
                run_start_time=run_start_time,
                record_length=samples_per_record,
                daq_chunk_duration=daq_chunk_duration,
                daq_overlap_chunk_duration=daq_overlap_chunk_duration,
                readout_threads=readout_threads,
                check_raw_record_overlaps=True,
            )
            log.info(f"Making {run_id}-{targets}")
            log.debug(f"With {strax_config}, n-cores {cores}")
            st.make(run_id, targets, allow_multiple=True, config=strax_config, max_workers=cores)

            if len(post_processing):
                for post_target in post_processing:
                    if post_target not in st._plugin_class_registry:
                        log_warning(
                            f"Trying to make unknown data type {post_target}",
                            priority="error",
                            run_id=run_id,
                        )
                        continue
                    elif not st.is_stored(run_id, post_target):
                        log.info(f"Making {post_target}")
                        st.make(
                            run_id,
                            post_target,
                            config=strax_config,
                            progress_bar=True,
                            max_workers=cores,
                        )
                    else:
                        log.info(f"Not making {post_target}, it is already stored")

        if args.profile.lower() == "false":
            st_make()
        else:
            prof_file = f"run{run_id}_{args.profile}"
            if ".prof" not in prof_file:
                prof_file += ".prof"
            log.info(f"starting with profiler, saving as {prof_file}")
            with strax.profile_threaded(prof_file):
                st_make()
    except Exception as e:
        log.warning(f"Ran into {e} while processing {run_id}", exc_info=True)

        # Write exception to file, so bootstrax can read it
        exc_info = strax.formatted_exception()
        log.warning(f"Uploading traceback {exc_info}")
        with open(exception_tempfile, mode="w") as f:
            f.write(exc_info)
        os.makedirs("./bootstrax_exceptions", exist_ok=True)
        with open(f"./bootstrax_exceptions/{run_id}_exception.txt", mode="w") as f:
            f.write(exc_info)
        raise


def last_file_write_time(match_location: str) -> datetime:
    """Get the datetime in UTC of the last file that was written in raw-records*

    :param match_location: Location where runs are written to
    :return: TZ aware timestamp.

    """
    last_time = datetime.fromtimestamp(0).replace(tzinfo=pytz.utc)
    matched_folders = glob(match_location)
    for folder in matched_folders:
        # check last three files only
        chunk_files = sorted(glob(os.path.join(folder, "*")))[-3:]
        for chunk in chunk_files:
            # Check that we did not rename this file since the glob above
            if os.path.exists(chunk):
                chunk_write_time = datetime.fromtimestamp(os.stat(chunk).st_mtime).replace(
                    tzinfo=pytz.utc
                )
                last_time = max(last_time, chunk_write_time)
    return last_time


def process_run(rd, send_heartbeats=args.production):
    log.info(f"Starting processing of run {rd['number']}")
    if rd is None:
        raise RuntimeError("Pass a valid rundoc, not None!")
    elif args.production and rd.get("bootstrax", {}).get("state", None) == "done":
        raise RuntimeError(f'{rd["number"]} is done already, do not make a mass')

    # Shortcuts for failing
    class RunFailed(Exception):
        pass

    def fail(reason, **kwargs):
        if args.production:
            fail_run(rd, reason, **kwargs)
        else:
            log.warning(reason)
        raise RunFailed

    try:
        try:
            run_id = "%06d" % rd["number"]
        except Exception as e:
            fail(f"Could not format run number: {str(e)}")

        if not args.production:
            # We are just testing let's assume its on the usual location
            loc = osp.join("/live_data/xenonnt/", run_id)
            # or use the test-dir:
            if not osp.exists(loc):
                loc = os.path.join("/live_data/xenonnt_bootstrax_test/", run_id)

        else:
            for dd in rd["data"]:
                if "type" not in dd:
                    fail("Corrupted data doc, found entry without 'type' field")
                if dd["type"] == "live":
                    break
                else:
                    fail("Non-live data already registered; untracked failure?")
            loc = osp.join(dd["location"], run_id)
        if not osp.exists(loc):
            fail(f"No live data at claimed location {loc}")

        run_strax_config = dict(run_id=run_id, input_dir=loc)

        if "daq_config" not in rd:
            fail("No daq_config in the rundoc!")
        try:
            # Fetch parameters from the rundoc. If not readable, let's use redax' default
            # values (that are hardcoded here).
            dq_conf = rd["daq_config"]
            to_read = (
                "processing_threads",
                "strax_chunk_length",
                "strax_chunk_overlap",
                "strax_fragment_payload_bytes",
                "compressor",
            )
            report_missing_config = [conf for conf in to_read if conf not in dq_conf]
            if report_missing_config:
                log_warning(
                    f'{", ".join(report_missing_config)} not in rundoc for '
                    f"{run_id}! Using default values.",
                    priority="warning",
                    run_id=run_id,
                )
            run_strax_config["readout_threads"] = dq_conf.get("processing_threads", None)
            run_strax_config["daq_chunk_duration"] = int(dq_conf.get("strax_chunk_length", 5) * 1e9)
            run_strax_config["daq_overlap_chunk_duration"] = int(
                dq_conf.get("strax_chunk_overlap", 0.5) * 1e9
            )
            # note that value in rd in bytes hence //2
            run_strax_config["samples_per_record"] = (
                dq_conf.get("strax_fragment_payload_bytes", 220) // 2
            )
            run_strax_config["compressor"] = dq_conf.get("compressor", "lz4")
        except Exception as e:
            fail(f"Could not find {to_read} in rundoc: {str(e)}")

        if run_strax_config["readout_threads"] is None:
            fail(f"Run doc for {run_id} has no readout thread count info")

        # Remove any previous processed data
        # If we do not do this, strax will just load this instead of
        # starting a new processing
        if args.production:
            clean_run(mongo_id=rd["_id"])
        else:
            clean_run_test_data(run_id)

        # Remove any temporary exception info from previous runs
        if osp.exists(exception_tempfile):
            os.remove(exception_tempfile)

        if not args.production and "bootstrax" not in rd:
            # Bootstrax does not register in non-production mode
            pass
        try:
            run_strax_config["run_start_time"] = (
                rd["start"].replace(tzinfo=timezone.utc).timestamp()
            )
        except Exception as e:
            fail(f"Could not find start in datetime.datetime object: {str(e)}")

        run_strax_config.update(infer_target(rd))
        run_strax_config.update(infer_mode(rd))
        run_strax_config["debug"] = args.debug
        strax_proc = multiprocessing.Process(target=run_strax, kwargs=run_strax_config)

        t0 = now()
        info = dict(started_processing=t0)
        strax_proc.start()

        while True:
            if send_heartbeats:
                to_report = [
                    "run_id",
                    "targets",
                    "cores",
                    "max_messages",
                    "timeout",
                    "post_processing",
                ]
                update = {k: v for k, v in run_strax_config.items() if k in to_report}
                send_heartbeat(update)
            ec = strax_proc.exitcode
            if ec is None:
                # fail(bla) raises RunFailed so no need for elifs. Make sure to
                # kill main process before raising errors in fail(bla)

                # Fail because we are taking a very long time to process
                if t0 < now(-timeouts["max_processing_time"]):
                    kill_process(strax_proc.pid)
                    fail(f"Processing took longer than {timeouts['max_processing_time']} sec")

                # Fail because we are taking a very long time to process
                # despite the run having ended
                endtime = get_end_time(rd)
                if (
                    endtime is not None
                    and endtime.replace(tzinfo=pytz.utc)
                    < now(-timeouts["max_processing_time_after_completion"])
                    and t0 < now(-timeouts["max_processing_time_after_completion"])
                ):
                    kill_process(strax_proc.pid)
                    fail(
                        "After the run ended, processing did not succeed in "
                        f"{timeouts['max_processing_time_after_completion']} sec."
                    )

                # Fail because for some reason, we are not writing any new files to disk.
                if t0 < now(-timeouts["max_no_write_time"]) and (
                    last_write := last_file_write_time(os.path.join(output_folder, f"{run_id}*"))
                ) < now(-timeouts["max_no_write_time"]):
                    kill_process(strax_proc.pid)
                    fail(
                        f"No data written for {timeouts['max_no_write_time']} sec. Last write is"
                        f" from {last_write}"
                    )

                if args.production:
                    set_run_state(rd, "busy", **info)
                time.sleep(timeouts["check_on_strax"])
                log.info(f"Still processing run {run_id}. PID:{strax_proc.pid}")
                continue

            elif ec == 0:
                log.info(f"Strax done on run {run_id}, performing basic data quality check")
                if args.ignore_checks:
                    # I hope you know what you are doing, we are not going to
                    # do any of the checks below.
                    # Make sure to fetch the latest rundoc
                    rd = get_run(mongo_id=rd["_id"])
                else:
                    log.info(f"Open metadata")
                    try:
                        # Sometimes we have only he channels or mv channels,
                        # try loading one until we get one with chunks.
                        for rr_type in (
                            "raw_records",
                            "raw_records_he",
                            "raw_records_mv",
                            "raw_records_nv",
                        ):
                            md = st.get_metadata(run_id, rr_type)
                            if len(md["chunks"]) and (
                                "first_time" in md["chunks"][0]
                                and "last_endtime" in md["chunks"][0]
                            ):
                                log.info(f"Using {rr_type}-metadata")
                                break
                    except Exception:
                        fail(
                            "Processing succeeded, but metadata not readable",
                            error_traceback=traceback.format_exc(),
                        )
                    if not len(md["chunks"]):
                        fail("Processing succeeded, but no chunks were written!")

                    if any(
                        (chunk["n"] and "filesize" in chunk and not chunk["filesize"])
                        for chunk in md["chunks"]
                    ):
                        # E.g. you tried compressing >2 GB chunk using blosc
                        fail("At least one chunk failed writing!")

                    log.info(f"Check that run has ended")
                    rd = get_run(mongo_id=rd["_id"])
                    end_time = rd.get("end")
                    if end_time is None:
                        fail("Processing succeeded, but run hasn't yet ended!")

                    log.info(f"Check the processing time of the run")
                    # Check that the data written covers the run
                    # (at least up to some fudge factor)
                    # Since chunks can be empty, and we don't want to crash,
                    # this has to be done with some care...
                    # Lets assume some ridiculous timestamp (in ns): 10e9*1e9
                    t_covered = timedelta(
                        seconds=(
                            max([x.get("last_endtime", 0) for x in md["chunks"]])
                            - min([x.get("first_time", 10e9 * 1e9) for x in md["chunks"]])
                        )
                        / 1e9
                    )
                    log.info(f"Compute runtime")
                    run_duration = end_time - rd["start"]
                    if not (0 < t_covered.total_seconds() < float("inf")):
                        fail(f"Processed data covers {t_covered} sec")
                    if not (
                        timedelta(seconds=-max_timetamp_diff)
                        < (run_duration - t_covered)
                        < timedelta(seconds=max_timetamp_diff)
                    ):
                        fail(
                            f"Processing covered {t_covered.total_seconds()}, "
                            f"but run lasted {run_duration.total_seconds()}!"
                        )
                    if args.production:
                        # Only check rundoc for files written in production mode
                        log.info(f"Check files saved")
                        if not all_files_saved(rd):
                            fail("Not all files in the rundoc for this run are saved")

                log.info(f"Run {run_id} processed successfully")
                if args.production:
                    set_run_state(rd, "done", **info)

                    set_status_finished(rd)

                    if args.delete_live:
                        delete_live_data(rd, loc)
                break

            else:
                # This is just the info that we're starting
                # exception retrieval. The actual error comes later.
                log.info(f"Failure while processing run {run_id}")
                if osp.exists(exception_tempfile):
                    with open(exception_tempfile, mode="r") as f:
                        exc_info = f.read()
                    if not exc_info:
                        exc_info = "[No exception info known, exception file was empty?!]"
                else:
                    exc_info = "[No exception info known, exception file not found?!]"
                fail(
                    f"Strax exited with exit code {ec}.",
                    error_traceback=f"Exception info: {exc_info}",
                )
    except RunFailed:
        return


##
# Cleanup
##


def clean_run(*, mongo_id=None, number=None, force=False):
    """Removes all data on this host associated with a run that was previously registered in the run
    db.

    Does NOT remove temporary folders, nor data that isn't registered to the run db.

    """
    # We need to get the full data docs here, since I was too lazy to write
    # a surgical update below
    rd = get_run(mongo_id=mongo_id, number=number, full_doc=True)
    have_live_data = False
    for dd in rd["data"]:
        if dd["type"] == "live":
            have_live_data = True
            break
    for ddoc in rd["data"]:
        if "host" in ddoc and ddoc["host"] == hostname:
            loc = ddoc["location"]
            if not force and not have_live_data and "raw_records" in ddoc["type"]:
                log.info(
                    f"prevent {loc} from being deleted. The live_data has already been removed"
                )
            elif os.path.exists(loc):
                log.info(f"delete data at {loc}")
                _delete_data(rd, loc, ddoc["type"])
            else:
                loc = loc + "_temp"
                log.info(f"delete data at {loc}")
                _delete_data(rd, loc, ddoc["type"])

    # Also wipe the online_monitor if there is any
    run_db["online_monitor"].delete_many({"number": int(rd["number"])})


def clean_run_test_data(run_id):
    """Clean the data in the test_data_folder associated with this run_id."""
    for folder in os.listdir(test_data_folder):
        if run_id in folder:
            log.info(f"Cleaning {test_data_folder + folder}")
            shutil.rmtree(test_data_folder + folder)


def cleanup_db():
    """Find various pathological runs and clean them from the db.

    Also cleans the bootstrax collection for stale entries

    """
    # Check mongo connection
    ping_dbs()

    log.info("Checking for bad stuff in database")

    # Check for all the ebs if their last state message is not longer
    # ago than the time we assume that the eb is dead.
    for eb_i in range(6):
        bd = bs_coll.find_one(
            {"host": f"eb{eb_i}.xenon.local"}, sort=[("time", pymongo.DESCENDING)]
        )
        if (
            bd
            and bd["time"].replace(tzinfo=pytz.utc) < now(-timeouts["bootstrax_presumed_dead"])
            and bd["state"] is not dead_state
        ):
            bs_coll.find_one_and_update({"_id": bd["_id"]}, {"$set": {"state": dead_state}})

    # Runs that say they are 'considering' or 'busy' but nothing happened for a while
    for state, timeout in [
        ("considering", timeouts["max_considering_time"]),
        ("busy", timeouts["max_busy_time"]),
    ]:
        while True:
            send_heartbeat()
            rd = consider_run(
                {"bootstrax.state": state, "bootstrax.time": {"$lt": now(-timeout)}},
                return_new_doc=False,
            )
            if rd is None:
                break
            fail_run(
                rd,
                f"Host {rd['bootstrax']['host']} said it was {state} "
                f"at {rd['bootstrax']['time']}, but then didn't get further; "
                "perhaps it crashed on this run or is still stuck?",
            )

    # Runs for which, based on the run doc alone, we can tell they are in a bad state
    # Mark them as failed.
    failure_queries = [
        (
            {"bootstrax.state": "done", "end": None},
            "Bootstrax state was done, but run did not yet end",
        ),
        (
            {
                "bootstrax.state": "done",
                "data": {"$not": {"$elemMatch": {"type": {"$ne": "live"}}}},
            },
            "Bootstrax state was done, but no processed data registered",
        ),
    ]

    for query, failure_message in failure_queries:
        while True:
            send_heartbeat()
            rd = consider_run(query)
            if rd is None:
                break
            fail_run(rd, failure_message.format(**rd))

    # Abandon runs which we already know are so bad that
    # there is no point in retrying them
    abandon_queries = [
        (
            {
                "tags.name": "abandon",
                "bootstrax.state": "done",
                "start": {"$gt": now(-timeouts["abandoning_allowed"])},
            },
            "Run has an 'abandon' tag",
        ),
        (
            {"tags.name": "abandon", "bootstrax.state": "failed"},
            "Run has an 'abandon' tag and was failing",
        ),
    ]

    for query, failure_message in abandon_queries:
        failure_message += " -- run has been abandoned"
        while True:
            send_heartbeat()
            rd = consider_run(query)
            if rd is None:
                break
            fail_run(rd, failure_message.format(**rd))
            abandon(mongo_id=rd["_id"])


def main():
    if not args.undying:
        run()
    else:
        while True:
            try:
                run()
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception as fatal_error:
                log.error(
                    f"Fatal warning:\tran into {fatal_error}. Try "
                    "logging error and restart bootstrax"
                )
                try:
                    log_warning(f"Fatal warning:\tran into {fatal_error}", priority="error")
                except Exception as warning_error:
                    log.error(f"Fatal warning:\tcould not log {warning_error}")
                # This usually only takes a minute or two
                time.sleep(60)
                log.warning("Restarting run loop")


if __name__ == "__main__":
    main()



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventNearestTriggering(strax.Plugin):
    """Time difference and properties of the nearest triggering peaks of main peaks of events."""

    __version__ = "0.0.0"
    depends_on = ("event_basics", "peak_basics", "peak_nearest_triggering")
    provides = "event_nearest_triggering"
    save_when = strax.SaveWhen.EXPLICIT

    def infer_dtype(self):
        dtype = []
        common_descr = "of the nearest triggering peak on the"
        for main_peak, main_peak_desc in zip(["s1_", "s2_"], ["main S1", "main S2"]):
            for direction in ["left", "right"]:
                dtype += [
                    (
                        (
                            f"time difference {common_descr} {direction} of {main_peak_desc} [ns]",
                            f"{main_peak}{direction}_dtime",
                        ),
                        np.int64,
                    ),
                    (
                        (
                            f"time {common_descr} {direction} of {main_peak_desc} [ns]",
                            f"{main_peak}{direction}_time",
                        ),
                        np.int64,
                    ),
                    (
                        (
                            f"endtime {common_descr} {direction} of {main_peak_desc} [ns]",
                            f"{main_peak}{direction}_endtime",
                        ),
                        np.int64,
                    ),
                    (
                        (
                            f"center_time {common_descr} {direction} of {main_peak_desc} [ns]",
                            f"{main_peak}{direction}_center_time",
                        ),
                        np.int64,
                    ),
                    (
                        (
                            f"type {common_descr} {direction} of {main_peak_desc}",
                            f"{main_peak}{direction}_type",
                        ),
                        np.int8,
                    ),
                    (
                        (
                            f"proximity_score {common_descr} {direction} of {main_peak_desc}",
                            f"{main_peak}{direction}_proximity_score",
                        ),
                        np.float32,
                    ),
                    (
                        (
                            f"n_competing_left {common_descr} {direction} of {main_peak_desc}",
                            f"{main_peak}{direction}_n_competing_left",
                        ),
                        np.int32,
                    ),
                    (
                        (
                            f"n_competing {common_descr} {direction} of {main_peak_desc}",
                            f"{main_peak}{direction}_n_competing",
                        ),
                        np.int32,
                    ),
                    (
                        (
                            f"area {common_descr} {direction} of {main_peak_desc} [PE]",
                            f"{main_peak}{direction}_area",
                        ),
                        np.float32,
                    ),
                ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.dtype)

        strax.set_nan_defaults(result)

        # 1. Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            res_i = result[event_i]
            # Fetch the features of main S1 and main S2
            for idx, main_peak in zip([event["s1_index"], event["s2_index"]], ["s1_", "s2_"]):
                if idx >= 0:
                    for direction in ["left", "right"]:
                        for field in [
                            "dtime",
                            "time",
                            "endtime",
                            "center_time",
                            "type",
                            "proximity_score",
                            "n_competing_left",
                            "n_competing",
                            "area",
                        ]:
                            res_i[f"{main_peak}{direction}_{field}"] = sp[f"{direction}_{field}"][
                                idx
                            ]

        # 2. Set time and endtime for events
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)
        return result



from immutabledict import immutabledict
import strax
import straxen

from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.records.records import PulseProcessing, pulse_count_dtype

export, __all__ = strax.exporter()


@export
class PulseProcessingHighEnergy(PulseProcessing):
    __doc__ = HE_PREAMBLE + (PulseProcessing.__doc__ or "")
    __version__ = "0.0.1"
    provides = ("records_he", "pulse_counts_he")
    data_kind = {k: k for k in provides}
    rechunk_on_save = immutabledict(records_he=False, pulse_counts_he=True)
    depends_on = "raw_records_he"
    compressor = "zstd"
    child_plugin = True
    save_when = strax.SaveWhen.TARGET

    n_he_pmts = straxen.URLConfig(
        track=False, default=752, infer_type=False, help="Maximum channel of the he channels"
    )

    record_length = straxen.URLConfig(
        default=110, track=False, type=int, help="Number of samples per raw_record"
    )

    hit_min_amplitude_he = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://hit_thresholds"
            "?as_list=True"
            "&sort=pmt"
            "&attr=value"
            "&detector=tpc_he"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
        ),
        track=True,
        infer_type=False,
        help="Minimum hit amplitude in ADC counts above baseline. "
        "Specify as a tuple of length n_tpc_pmts, or a number,"
        'or a string like "pmt_commissioning_initial" which means calling'
        "hitfinder_thresholds.py",
    )

    def infer_dtype(self):
        dtype = dict()
        dtype["records_he"] = strax.record_dtype(self.record_length)
        dtype["pulse_counts_he"] = pulse_count_dtype(self.n_he_pmts)
        return dtype

    def setup(self):
        self.hev_enabled = False

        # FIXME: This looks hacky. Maybe find a better way?
        self.config["n_tpc_pmts"] = self.config["n_he_pmts"]

        self.hit_thresholds = self.hit_min_amplitude_he

    def compute(self, raw_records_he, start, end):
        result = super().compute(raw_records_he, start, end)
        return dict(records_he=result["records"], pulse_counts_he=result["pulse_counts"])



import numpy as np
from numba import njit

import strax
import straxen


class PeakSEScore(strax.OverlapWindowPlugin):
    """This plugin is designed to calculate a score based on the position relations to single
    electrons.

    References:
        * v0.2.0: xenon:xenonnt:analysis:hot_spot_cut_summary
        * v0.4.0: xenon:xenonnt:ac:sr1:hotspot_veto_cut:wimp_roi
        * position resolution: xenon:xenonnt:svetter:data_driven_resolution_sr1a0_vs_sr1a1

    """

    __version__ = "0.4.0"
    depends_on = ("peak_basics", "peak_positions")
    provides = "peak_se_score"
    save_when = strax.SaveWhen.EXPLICIT

    dtype = strax.time_fields + [
        (
            ("Sum of the PDF of the SE position nearby probability.", "se_score"),
            np.float32,
        ),
    ]

    se_time_search_window_left = straxen.URLConfig(
        default=5e9, type=int, help="SEs time window searching window, left side [ns]"
    )

    se_time_search_window_right = straxen.URLConfig(
        default=0,
        type=int,
        help="Time searching window right side extension, one max drift time [ns]",
    )

    se_selection_area_roi = straxen.URLConfig(
        default=[10, 80],
        type=list,
        help="Area range for single electron selection.[PE]",
    )

    se_selection_width_roi = straxen.URLConfig(
        default=[80, 700],
        type=list,
        help="Area range for single electron selection.[PE]",
    )

    para_a_naive = straxen.URLConfig(
        default=1,
        help="Place holder for Parameter A in the position resolution function.",
    )

    para_b_naive = straxen.URLConfig(
        default=1,
        help="Place holder for Parameter B in the position resolution function.",
    )

    def get_window_size(self):
        # This method is required by the OverlapWindowPlugin class
        return 2 * (self.se_time_search_window_left + self.se_time_search_window_right)

    def setup(self):
        self._para_a = self.para_a_naive
        self._para_b = self.para_b_naive

    def select_se(self, peaks):
        """Function which select single electrons from peaks.

        :param peaks: peaks data contains single electrons.
        :return: single electron peaks data

        """
        mask = peaks["type"] == 2
        mask &= (peaks["area"] > self.se_selection_area_roi[0]) & (
            peaks["area"] < self.se_selection_area_roi[1]
        )
        mask &= (peaks["range_50p_area"] > self.se_selection_width_roi[0]) & (
            peaks["range_50p_area"] < self.se_selection_width_roi[1]
        )
        mask &= peaks["area_fraction_top"] != 0
        return mask

    @staticmethod
    @njit
    def get_se_count_and_pdf_sum(peaks, se_peaks, se_indices, para_a, para_b):
        """Function to calculate the single electron (SE) score for each peak.

        This function calculates the probability of nearby SE events for each peak
        based on their spatial and temporal proximity. The probability is calculated
        using a Gaussian distribution, where the standard deviation is the combined
        position resolution of the peak and the SE event.

        Parameters:
        :peaks (np.ndarray): Array of peaks.
        :se_peaks (np.ndarray): Array of single electron peaks.
        :se_indices (np.ndarray): Array of indices for SE that are
        in temporal proximity to each peak.
        :para_a (float): Parameter 'a' used in the calculation of position resolution.
        :para_b (float): Parameter 'b' used in the calculation of position resolution.

        :return: np.ndarray: Array of probabilities of nearby SE events for each peak.

        """
        se_nearby_probability = np.zeros(len(peaks))
        for index, peak_i in enumerate(peaks):
            indices = se_indices[index]
            se_in_time = se_peaks[indices[0] : indices[1]]
            peak_area_top = peak_i["area"] * peak_i["area_fraction_top"]
            se_area_top = se_in_time["area"] * se_in_time["area_fraction_top"]
            peak_position_resolution = para_a / np.sqrt(peak_area_top) + para_b
            se_position_resolution = para_a / np.sqrt(se_area_top) + para_b
            combined_position_resolution = np.sqrt(
                se_position_resolution**2 + peak_position_resolution**2
            )
            d_squre = (se_in_time["x"] - peak_i["x"]) ** 2 + (se_in_time["y"] - peak_i["y"]) ** 2
            # 2D Gaussian with zero off-diagonal covairance matirx elements
            constant = 1 / (2 * np.pi * combined_position_resolution**2)
            exponent = np.exp(-1 / 2 * (d_squre / combined_position_resolution**2))
            _se_nearby_prob = constant * exponent
            se_nearby_probability[index] = np.sum(_se_nearby_prob)
        return se_nearby_probability

    def compute_se_score(self, peaks, _peaks):
        """Function to calculate the SE score for each peak."""
        # select single electrons
        mask = self.select_se(peaks)
        se_peaks = peaks[mask]
        # get single electron indices in peak vicinity
        split_peaks = np.zeros(len(_peaks), dtype=strax.time_fields)
        split_peaks["time"] = _peaks["center_time"] - self.se_time_search_window_left
        split_peaks["endtime"] = _peaks["center_time"] + self.se_time_search_window_right
        split_result = strax.touching_windows(se_peaks, split_peaks)
        # get se score
        _se_nearby_probability = self.get_se_count_and_pdf_sum(
            _peaks,
            se_peaks,
            split_result,
            self._para_a,
            self._para_b,
        )
        return _se_nearby_probability

    def compute(self, peaks):
        # sort peaks by center_time
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = peaks[argsort].copy()
        mask_nan = np.isnan(_peaks["x"]) | np.isnan(_peaks["y"])
        # prepare output
        se_nearby_probability = np.zeros(len(peaks))
        _se_nearby_probability = np.full(len(peaks), np.nan)
        # calculate SE Score
        _se_nearby_probability[~mask_nan] = self.compute_se_score(peaks, _peaks[~mask_nan])
        # sort back to original order
        se_nearby_probability[argsort] = _se_nearby_probability
        return dict(
            time=peaks["time"],
            endtime=strax.endtime(peaks),
            se_score=se_nearby_probability,
        )



"""
Restrax: Rechunking live data
=============================================
How to use
----------------
    <activate conda environment>
    restrax --production

----------------

For more info, see the documentation:
https://straxen.readthedocs.io/en/latest/scripts.html
"""

__version__ = "0.3.1"

import argparse
import logging
import os
import typing

import immutabledict
import socket
import shutil
import time
import numpy as np
import strax
import straxen
import threading
import daqnt
import fnmatch
import typing as ty
from straxen import daq_core
from memory_profiler import memory_usage
import glob
from ast import literal_eval
from straxen.daq_core import now


def parse_args():
    parser = argparse.ArgumentParser(description="XENONnT rechunking manager")
    parser.add_argument(
        "--production",
        action="store_true",
        help="Run restrax in production mode, otherwise run in a test mode.",
    )
    parser.add_argument(
        "--ignore_checks",
        action="store_true",
        help="Do not use! Skip checks before changing documents, there be dragons.",
    )
    parser.add_argument(
        "--input_folder",
        type=str,
        default=daq_core.pre_folder,
        help="Where to read the data from, should only be used when testing",
    )
    parser.add_argument(
        "--max_threads",
        type=int,
        default=2,
        help="max number of threads to simultaneously work on one run",
    )
    parser.add_argument(
        "--skip_compression",
        nargs="*",
        default=["*event*"],
        help='skip recompression of any datatype that fnmatches. For example: "*event* *peak*"',
    )
    # deep_compare is meant only for testing
    parser.add_argument(
        "--deep_compare",
        action="store_true",
        help="Open all the data of the old and the new format and check that they are the same",
    )
    parser.add_argument(
        "--recompress_min_chunks",
        type=int,
        default=2,
        help="Only bother with doing the recompression if there are more than this many chunks",
    )
    parser.add_argument(
        "--bypass_mode",
        action="store_true",
        help="Stop recompression and just rename folders. Use with care!",
    )
    actions = parser.add_mutually_exclusive_group()
    actions.add_argument("--undying", action="store_true", help="Except any error and ignore it")
    actions.add_argument("--process", type=int, help="Handle a single run")
    args = parser.parse_args()
    if args.input_folder != daq_core.pre_folder and args.production:
        raise ValueError(
            "Thou shall not pass, don't upload files from non production"
            " folders, what are you thinking you're doing?!!"
        )
    return args


def main():
    args = parse_args()
    restrax = ReStrax(args=args)

    while True:
        try:
            run_restrax(restrax, args)
            if args.process:
                break
        except (KeyboardInterrupt, SystemExit):
            break
        except Exception as fatal_error:
            restrax.log.error(
                f"Fatal warning: ran into {fatal_error}. Trying to log error and restart ReStrax."
            )
            try:
                restrax.log_warning(
                    f"Fatal warning: ran into {fatal_error}",
                    priority="error",
                )
            except Exception as warning_error:
                restrax.error(f"Fatal warning: could not log {warning_error}")

            if not args.undying:
                raise

            restrax.log.warning("Restarting main loop after 60 seconds due to fatal error.")
            time.sleep(60)


def run_restrax(restrax, args):
    """Run the infinite loop of ReStrax."""
    restrax.infinite_loop(close=bool(args.process))


class ReStrax(daq_core.DataBases):
    """
    Restrax: rechunking the data from bootstrax and prepare it for admix/rucio
    """

    # Typing is important because of the overwrite_settings
    log: logging.Logger
    max_threads: int
    ignore_checks: bool
    skip_compression: bool
    process: ty.Union[str, int]
    deep_compare: bool
    recompress_min_chunks: int
    bypass_mode: bool

    nap_time: ty.Union[int, float] = 300  # s
    nap_time_short: ty.Union[int, float] = 5  # s

    folders: ty.Mapping = immutabledict.immutabledict(
        production_out=daq_core.output_folder, test_out="/data/test_processed"
    )

    raw_record_types: ty.Iterable = (
        "raw_records",
        "raw_records_nv",
        "raw_records_mv",
        "raw_records_he",
        "records",
        "records_nv",
        "records_mv",
    )

    exclude_modes: ty.Iterable = (
        "pmtgain",
        "pmtap",
        "exttrig",
        "noise",
        "nVeto_LASER_calibration",
        "mv_diffuserballs",
        "mv_fibres",
        "mv_darkrate",
    )

    # If a dataset is larger than this many bytes, only compare a single field (time)
    large_data_compare_threshold: ty.Union[int, float] = int(5e9)

    # If the total data rate is lower than this, use bz2 for the compression of raw records.
    is_heavy_rate_mbs: ty.Union[int, float] = 50  # MB/s

    # Default compressors / target sizes. Fn matches the datatype to the keys in the dict to get the
    target_compressor: ty.Mapping = immutabledict.immutabledict(
        {
            "peak*": "zstd",
            "_raw_record_compressor_light": "bz2",
            "_raw_record_compressor_heavy": "zstd",
            "_other": None,  # use the default compressor
        }
    )
    target_size: ty.Mapping = immutabledict.immutabledict(
        {
            "peak*": 1_500,
            "_other": 1_500,
            "_raw_records": 5_000,
        }
    )

    # Prevent using more than this much MB when doing data validation
    max_compare_buffer_mb: ty.Union[float, int] = 5000

    # Try recompressing this many times, inform the log database otherwise about failures
    max_tries: int = 5

    # Settings for rechunking
    parallel: ty.Optional[ty.Union[str, bool]] = True  # (either True, False or 'process')
    max_workers: int = 4

    # Timeout for the mailboxes responsible for saving the data (including compression!)
    _saver_timeout: ty.Union[float, int] = 3600  # s

    def __init__(self, args):
        super().__init__(production=args.production)

        self.hostname = socket.getfqdn()
        self._set_logger()
        # Get from the database unless testing
        self.read_from = None if args.production else args.input_folder
        self.write_to = (
            self.folders["production_out"] if args.production else self.folders["test_out"]
        )
        self.max_threads = args.max_threads
        self.ignore_checks = args.ignore_checks
        self.skip_compression = args.skip_compression
        self.process = args.process
        self.deep_compare = args.deep_compare
        self.recompress_min_chunks = args.recompress_min_chunks
        self.bypass_mode = args.bypass_mode
        self.overwrite_settings()

    def infinite_loop(self, close=False) -> None:
        """Core of restrax, recompress the data followed by several validation steps."""
        while True:
            self.overwrite_settings()
            run_doc = self.find_work()
            self.log.info("Start")
            if run_doc is None:
                if close:
                    return
                self.log.info("No work to do, sleep")
                self.take_a_nap()
                continue
            t0 = time.time()
            self.set_restrax_busy(run_doc)
            try:
                mem_for_doc = memory_usage(
                    (self.handle_run, (run_doc,)), max_iterations=1, interval=60
                )
            except Exception as exception:
                self.set_restrax_failed(run_doc, str(exception))
                self.log.error(f"Ran into {exception}")
                raise exception
            self.log.debug(f"Memory profiler says peak RAM usage was: {max(mem_for_doc):.1f} MB")
            self.log.debug(f"Took {(time.time() - t0) / 3600:.2f} h")
            self.set_restrax_done(run_doc, mem_for_doc)
            self.log.info("Loop done")
            if close:
                break

    def find_work(self, projection: ty.Optional[dict] = None) -> ty.Optional[dict]:
        """Get a list of documents to recompress and the associated rundoc.

        :param projection: optional, which fields from the run document to query
        :return: list of data documents, and the run-document that

        """
        if projection is None:
            projection = {k: 1 for k in "data number mode detectors rate".split()}
        if self.production:
            return self._find_production_work(projection)
        return self._find_testing_work(projection)

    def _find_testing_work(self, projection: dict) -> ty.Optional[dict]:
        """Find work from the pre-dir if we are testing."""
        folders = os.listdir(self.read_from)
        first_run = f"{int(self.process):06}"
        data_docs = []
        for folder in folders:
            if os.path.exists(os.path.join(self.write_to, folder)):
                # Don't do work twice
                continue
            if len(split := folder.split("-")) and len(split[0]) == 6:
                run_id, data_type, lineage = split

                if first_run is None:
                    first_run = run_id
                if run_id != first_run:
                    continue
                self.log.info(f"Do {folder}")
                data_docs.append(
                    {
                        "host": self.hostname,
                        "location": os.path.join(self.read_from, folder),
                        "type": data_type,
                        "linage_hash": lineage,
                    }
                )
        if not len(data_docs):
            return None
        run_doc = self.run_coll.find_one({"number": int(first_run)}, projection=projection)
        run_doc["data"] = data_docs
        return run_doc

    def _find_production_work(self, projection) -> ty.Optional[dict]:
        """Query the database for work."""
        query = {
            "status": "eb_finished_pre",
            "restrax": None,
            "bootstrax.state": "done",
            "bootstrax.host": self.hostname,
        }
        if self.process:
            query["number"] = int(self.process)
        kw = dict(sort=[("_id", -1)], projection=projection)
        run_doc = self.run_coll.find_one(query, **kw)

        if run_doc is None:
            self.log.info("No new work, looking for previously failed runs")
            # Look for work which we tried before (and has a restrax field)
            query.pop("restrax")
            query.update(
                {"restrax.n_tries": {"$lt": self.max_tries + 1}, "restrax.state": {"$ne": "done"}}
            )
            run_doc = self.run_coll.find_one(query, **kw)

        if run_doc is not None:
            # Update data field in place
            run_doc["data"] = self._get_data_docs(run_doc)
        return run_doc

    def set_restrax_busy(self, run_doc: dict) -> None:
        self.log.info(f"Update/create restrax doc")
        if not self.production:
            return
        restrax_doc = {
            "n_tries": 1 if "restrax" not in run_doc else run_doc["restrax"]["n_tries"] + 1,
            "state": "busy",
            "started_processing": now(),
            "host": self.hostname,
            "ended": None,
            "memory_usage_mb": None,
        }
        self.run_coll.update_one(
            {"_id": run_doc["_id"]},
            {"$set": {"restrax": restrax_doc}},
        )

    def set_restrax_done(self, run_doc: dict, memory_usage_mb: list) -> None:
        """Update the rundoc with the restrax state."""
        update = {
            "restrax.state": "done",
            "restrax.ended": now(),
            "restrax.memory_usage_mb": {
                "avg": np.average(memory_usage_mb),
                "max": np.max(memory_usage_mb),
            },
        }
        if self.production:
            self.run_coll.update_one(
                {"_id": run_doc["_id"]},
                {"$set": update},
            )
        self.log.debug(f'Finished {run_doc["number"]}, set to {update}')

    def set_restrax_failed(self, run_doc: dict, reason: str) -> None:
        """Update the rundoc with the restrax fail state."""

        update = {
            "restrax.state": "failed",
            "restrax.ended": now(),
            "restrax.reason": reason,
        }

        if self.production:
            self.run_coll.update_one(
                {"_id": run_doc["_id"]},
                {"$set": update},
            )

            # Perform the $inc update to increment n_tries
            result_inc = self.run_coll.update_one(
                {"_id": run_doc["_id"]}, {"$inc": {"restrax.n_tries": 1}}
            )
            self.log.debug(f"Increment update result: {result_inc.raw_result}")

        self.log.debug(f'Fail {run_doc["number"]} with {update}')
        if run_doc.get("restrax", {}).get("n_tries", 0) >= self.max_tries:
            self.log_warning(
                f'Failed too many times for {run_doc["number"]}! '
                "I stop trying, manual help is needed."
            )

    def bypass_run(self, run_doc: dict) -> None:
        """Simply move all the data associate in the rundoc to the production folder."""
        # TODO: perhaps add some heavy try-except here to not allow interruption of
        # moving data to end up with splined datasets
        self.log.warning(f'Move {run_doc["number"]} to production. DO NOT INTERRUPT!')
        for data_doc in run_doc.get("data", []):
            self._bypass_for_data_doc(data_doc)

    def _bypass_for_data_doc(self, data_doc: dict) -> None:
        source = data_doc["location"]
        dest = self.renamed_path(source)
        if os.path.exists(dest):
            # Honestly don't know how this could happen, but we have to be carefull here
            # We are dealing with single copies, so this is a tricky operation.
            # Just to be sure we are not losing the only copy, let's make a backup
            # at the unregistered folder.
            move_to = self.renamed_path(source, _move_to=daq_core.non_registered_folder)
            self._move_dir(dest, move_to)
            message = f"{dest} already exists?! Backing up to {move_to}"
            self.log.error(message)
            self.log_warning(message)
        if not os.path.exists(source):
            # Complete chaos - this should never happen!
            message = f"Trying to move {source}->{dest} but {source} does not exist?!"
            self.log.error(message)
            self.log_warning(message)
            return
        self._move_dir(source, dest)

    def handle_run(self, run_doc: dict) -> None:
        """For a given batch of data_docs of a given run, do all the rechunking steps."""
        self.log.debug("start handle_run")
        data_docs = self._get_data_docs(run_doc)
        self.log.info(f'{run_doc["number"]} -> doing {len(data_docs)}')
        # self.run_software_veto(run_doc)

        # Split the work in files that we will compress and files that will skip compression
        compress_docs = [d for d in data_docs if not self.should_skip_compression(run_doc, d)]
        skip_docs = [d for d in data_docs if self.should_skip_compression(run_doc, d)]
        self.log.debug(
            f"Compressing {len(compress_docs)} docs and skipping (i.e. move) {len(skip_docs)} docs"
        )
        assert len(compress_docs) + len(skip_docs) == len(data_docs), "one and one is three?! "

        self.rechunk_docs(run_doc, compress_docs)
        self.do_checks(compress_docs)
        if self.deep_compare:
            self.validate_data(compress_docs)

        # In the final bookkeeping we move documents, move them only if all
        # checks above are good and update the rundoc in the finalize_execute
        for move_doc in skip_docs:
            self._bypass_for_data_doc(move_doc)

        self.finalize_execute(data_docs)

        # Only remove the data that we rechunked (there are now two copies),
        # the moved data is always just a single copy
        self.remove_old_docs(compress_docs)
        self.log.info(f'{run_doc["number"]} succes')

    def _get_data_docs(self, run_doc: dict):
        """Extract data doc from rundoc and sort by largest first."""
        # Filter data documents that are only on this host
        if run_doc.get("data") is None:
            return []
        data_docs = [
            data_doc for data_doc in run_doc["data"] if data_doc.get("host") == self.hostname
        ]
        storage_backend = strax.FileSytemBackend()
        size = lambda data_doc: sum(
            chunk.get("nbytes", 0)
            for chunk in storage_backend.get_metadata(data_doc["location"]).get("chunks", [dict()])
        )
        try:
            data_docs = sorted(data_docs, key=size, reverse=True)
        except Exception as exception:
            self.set_restrax_failed(run_doc, str(exception))
            self.log.error(f"Ran into {exception}")
            raise exception
        return data_docs

    def run_software_veto(self, run_doc: dict):
        """This is where we can add a software veto for specific runs."""
        raise NotImplementedError

    def rechunk_docs(self, run_doc: dict, data_docs: ty.List[dict]) -> None:
        """For each of the data documents, rechunk/recompress/move/check the data, if multi-
        threading is allowed, use that.

        :param run_doc: run document
        :param data_docs: list of data documents. NB! Will work on ALL these documents so these data
            documents don't have to be the same as the data field in the run_doc.

        """
        if self.max_threads > 1:
            threads: ty.List[ty.Any] = []
            for next_doc in data_docs:
                self._sleep_while_n_threads_alive(threads, self.max_threads - 1)
                thread = threading.Thread(
                    target=self._rechunk_per_doc,
                    args=(
                        run_doc,
                        next_doc,
                    ),
                )
                thread.start()
                threads.append(thread)
            # Wait until all threads have finished
            self._sleep_while_n_threads_alive(threads, 0)
        else:
            for ddoc in data_docs:
                self._rechunk_per_doc(run_doc, ddoc)

    def _rechunk_per_doc(self, run_doc: dict, data_doc: dict) -> None:
        """Do the rechunking document by document."""
        dir_in = data_doc["location"]
        dir_out = self.renamed_path(dir_in)

        if not os.path.exists(dir_in):
            raise FileNotFoundError(data_doc)

        self.log.info(f"Start {dir_in} -> {dir_out}")
        compressor, target_size_mb = self.get_compressor_and_size(run_doc, data_doc)

        self.log.debug(f"Running with parallel {self.parallel} and max_workers {self.max_workers}")
        # If ever performance is a too big an issue, we can easily change
        # `strax.rechunker` to work on multiple tartes at the same time
        # using the mailbox system.
        summary = strax.rechunker(
            source_directory=dir_in,
            dest_directory=self.write_to,
            replace=False,
            compressor=compressor,
            target_size_mb=target_size_mb,
            rechunk=True,
            parallel=self.parallel,
            max_workers=self.max_workers,
            _timeout=self._saver_timeout,
        )
        self.log.info(f"{dir_out} written {summary}")

    def _sleep_while_n_threads_alive(self, threads, n):
        while sum(t.is_alive() for t in threads) > n:
            self.take_a_nap(self.nap_time_short)

    def get_compressor_and_size(
        self, run_doc: dict, data_doc: dict
    ) -> ty.Tuple[str, ty.Optional[int]]:
        """For a given data document infer the desired compressor, and target size.

        :param run_doc: run document
        :param data_doc: data document

        """
        # This is where we might do some fancy coding
        dtype = data_doc["type"]
        if dtype in self.raw_record_types:
            rate = sum(
                detector.get("avg", 100) for detector in run_doc.get("rate", {"none": {}}).values()
            )
            _go_fast = (
                rate > self.is_heavy_rate_mbs
                or run_doc.get("restrax", {}).get("n_tries", 0) > self.max_tries // 2
            )
            compressor = (
                self.target_compressor.get("_raw_record_compressor_heavy", "bz2")
                if _go_fast
                else self.target_compressor.get("_raw_record_compressor_light", "zstd")
            )
            self.log.debug(
                f"Use {compressor}, we have to go fast: {_go_fast} (rate~{rate:.1f} MB/s)"
            )
            target_size_mb = self.target_size.get("_raw_records", 5000)
        else:
            # Extract the compressor from the settings
            compressor = self._fnmatch_from_doc(self.target_compressor, dtype)
            target_size_mb = self._fnmatch_from_doc(self.target_size, dtype)
            if target_size_mb is None:
                target_size_mb = 1_500

        self.log.debug(f"Setting {dtype}: {compressor}, {target_size_mb} MB")
        if compressor == "blosc" and target_size_mb > 1_500:
            # We really want to stay away from the 2 GB limit!
            self.log_warning(
                f"blosc fails compressing > 2 GB chunks, got instructed to do {target_size_mb} MB. "
                "Forcing a lower value!"
            )
            target_size_mb = 1_500
        return compressor, target_size_mb

    @staticmethod
    def _fnmatch_from_doc(some_dict, target_key):
        for target, value in some_dict.items():
            if fnmatch.fnmatch(target, target_key):
                return value
        return some_dict.get("_other", None)

    def should_skip_compression(self, run_doc: dict, data_doc: dict) -> bool:
        """Should we skip recompressing this data? For example if the data is already so small that
        it's not worth recompressing.

        :param run_doc: run document
        :param data_doc: data document

        """

        # Skipp all when in bypass mode - don't log messages for each data_doc separately
        if self.is_in_bypass_mode:
            return True

        data_type = data_doc["type"]
        run_mode = run_doc["mode"]
        if any([m in run_mode for m in self.exclude_modes]):
            self.log.debug(f"Skip mode {run_mode}")
            return True
        if data_type == "live_data":
            return True
        if any(fnmatch.fnmatch(data_type, delete) for delete in self.skip_compression):
            self.log.debug(f"Skip {data_type} -> matches skip_compression")
            return True
        n_chunks = len(
            strax.FileSytemBackend().get_metadata(data_doc["location"]).get("chunks", [])
        )
        if n_chunks <= self.recompress_min_chunks and data_type not in self.raw_record_types:
            # no need to recompress data if it's only one chunk
            self.log.debug(f"Skip {data_type} -> only {n_chunks} chunks")
            return True
        return False

    def do_checks(self, data_docs: ty.List[dict]) -> None:
        """Open the metadata of the old and the new file to check if everything is consistent and no
        exceptions were encountered during recompression.

        :param data_docs: list of data documents to rechunk

        """
        if self.ignore_checks:
            # One does not just venture into Mar a lago!
            return
        storage_backend = strax.FileSytemBackend()
        errors = []
        for data_doc in data_docs:
            dir_in = data_doc["location"]
            dir_out = self.renamed_path(dir_in)

            if not os.path.exists(dir_in):
                errors.append(f"{dir_in} does not exists.")
            if not os.path.exists(dir_out):
                errors.append(f"{dir_out} does not exists.")
            md_in = storage_backend.get_metadata(dir_in)
            md_out = storage_backend.get_metadata(dir_out)

            # Filesize is "bonus" metadata so it may not always be there.
            # Hence the .get(filesize, True)
            if any(
                (chunk.get("n", 0) and not chunk.get("filesize", True))
                for chunk in md_out["chunks"]
            ):
                # E.g. you tried compressing >2 GB chunk using blosc
                errors.append(f"For {dir_out}, at least one doc failed to write")

            if "exception" in md_out:
                errors.append("Writing error!")

            if sum(chunk["n"] for chunk in md_in["chunks"]) != sum(
                chunk["n"] for chunk in md_out["chunks"]
            ):
                errors.append("Rechunked data has fewer entries?!")
        if errors:
            locs = [d["location"] for d in data_docs]
            raise ValueError(f"Doc {locs} had errors: " + " and ".join(errors))
        self.log.info("All checks passed")

    def validate_data(self, data_docs: ty.List[dict]):
        for data_doc in data_docs:
            self._validate_data_per_doc(data_doc)

    def _validate_data_per_doc(self, data_doc: dict) -> None:
        """Open the data from the old and new source, compare to assert it's the same."""
        storage_backend = strax.FileSytemBackend()
        dir_in = data_doc["location"]
        dir_out = self.renamed_path(dir_in)

        md_in = storage_backend.get_metadata(dir_in)
        is_large_dset = (
            sum(chunk["nbytes"] for chunk in md_in["chunks"]) > self.large_data_compare_threshold
        )

        self.log.info(f"Checking {dir_in} vs {dir_out}. Is large {is_large_dset}")
        kw = dict(
            keep_column="time" if is_large_dset else None, progress_bar=len(md_in["chunks"]) > 4
        )
        data_in = self._get_data_from_dir(dir_in, **kw)  # type: ignore
        data_out = self._get_data_from_dir(dir_out, **kw)  # type: ignore
        errors = self._compare_data(data_in, data_out)
        if errors:
            raise ValueError(
                f"Data was not the same when comparing {dir_in} {dir_out}. See:\n"
                + "\n".join(errors)
            )
        self.log.debug("Compare done")

    @staticmethod
    def _compare_data(data_in: str, data_out: str):
        """Compare structured/simple numpy arrays."""
        errors = []
        if data_in.dtype.names:
            for field in data_in.dtype.names:
                if "str" in data_in.dtype[field].name:
                    # NB! The equal_nan compare does not work for string fields!
                    if not np.array_equal(data_in[field], data_out[field]):
                        errors.append(f'Error for "{field}"')
                elif not np.array_equal(data_in[field], data_out[field], equal_nan=True):
                    errors.append(f'Error for "{field}"')
        elif not np.array_equal(data_in, data_out, equal_nan=True):
            errors.append("Data not the same")
        return errors

    def _get_data_from_dir(
        self,
        folder: str,
        keep_column: typing.Optional[ty.Union[str, list]] = None,
        progress_bar: bool = False,
        _compare_sum: bool = False,
    ) -> np.ndarray:
        """Load data from a specified folder.

        :param folder: absolute path to folder
        :param keep_column: columns to keep from structured array (if None, get all)
        :param progress_bar: show a tqdm progressbar
        :param _compare_sum: compare only the sum of each field per chunk to prevent keeping too
            much memory occoupied. Will be overwritten for too large arrays.
        :return: concatenated numpy array

        """
        meta_data = strax.FileSytemBackend()._get_metadata(folder)
        files = strax.utils.tqdm(
            sorted(glob.glob(os.path.join(folder, "*")))[:-1],
            desc=folder.split("/")[-1],
            disable=not progress_bar,
        )

        dtype = literal_eval(meta_data["dtype"])

        if keep_column:
            res_dtype = [d for d in dtype if d[0][1] in strax.to_str_tuple(keep_column)]
            if len(res_dtype) == 1:
                res_dtype = res_dtype[0][-1]
        else:
            res_dtype = dtype

        data_len = sum(d["n"] for d in meta_data["chunks"])

        if (
            np.zeros(1, dtype=res_dtype).nbytes * data_len / 1e6 > self.max_compare_buffer_mb
            or _compare_sum
        ):
            if not _compare_sum:
                self.log.info(
                    f"Memory footprint for loading {folder} would be too large, only compare sum."
                )
            # Larger than 5 GB, only computing total sum of fields
            _compare_sum = True
            # We might get integer overflows, but that shouldn't be much of a problem
            # as long as we get the same number of overflows for both datasets
            result = np.zeros(1, dtype=res_dtype)
        else:
            result = np.zeros(data_len, dtype=res_dtype)

        n_seen = 0
        for file in files:
            chunk = strax.io.load_file(file, compressor=meta_data["compressor"], dtype=dtype)

            if _compare_sum and keep_column:
                result += np.sum(chunk[keep_column])
                continue
            elif _compare_sum:
                for field in result.dtype.names:
                    result[field] += np.sum(chunk[field])
                continue

            new = chunk
            if keep_column:
                new = chunk[keep_column].copy()

            result[n_seen : n_seen + len(new)] = new
            n_seen += len(new)
        return result

    def finalize_execute(self, data_docs: ty.List[dict]) -> None:
        """Update the database according to the operation and update the metadata.

        :param data_docs: list of data documents that have been rechunk
        :return:

        """
        # Maybe could merge this with do checks? -> Avoid opening metadate twice?
        # Then again, that is SO minor in the grand scheme of things,
        # that I just leave it like this for the moment
        if not self.production or not len(data_docs):
            return
        storage_backend = strax.FileSytemBackend()
        for data_doc in data_docs:
            if self.hostname != data_doc["host"]:
                continue
            dir_out = self.renamed_path(data_doc.get("location", ""))

            new_metadata = storage_backend.get_metadata(dir_out)
            run_id = new_metadata["run_id"]
            chunk_mb = [chunk["nbytes"] / 1e6 for chunk in new_metadata["chunks"]]

            if not self.production:
                continue

            self.log.info(f"Update data doc {dir_out}")
            self.run_coll.update_one(
                {
                    "number": int(run_id),
                    "data": {
                        "$elemMatch": {"location": data_doc["location"], "host": data_doc["host"]}
                    },
                },
                {
                    "$set": {
                        "data.$.location": dir_out,
                        "data.$.file_count": len(os.listdir(dir_out)),
                        "data.$.meta.strax_version": strax.__version__,
                        "data.$.meta.straxen_version": straxen.__version__,
                        "data.$.meta.size_mb": int(np.sum(chunk_mb)),
                        "data.$.meta.avg_chunk_mb": int(np.average(chunk_mb)),
                        "data.$.meta.lineage_hash": new_metadata["lineage_hash"],
                        "data.$.meta.compressor": new_metadata["compressor"],
                    }
                },
            )
        # Mark as ready for upload such that admix can take over from here
        if self.production:
            self.run_coll.update_one(
                {"number": int(run_id)}, {"$set": {"status": "eb_ready_to_upload"}}
            )
        self.log.info("Rundoc updated")

    def remove_old_docs(self, done_data_docs: ty.List[dict]):
        for data_doc in done_data_docs:
            loc = data_doc.get("location", "??")
            assert "pre_processed" in loc
            self._remove_dir(loc)

    def take_a_nap(self, dt: ty.Optional[int] = None):
        time.sleep(dt if dt is not None else self.nap_time)

    def renamed_path(self, path: str, _move_to: ty.Optional[str] = None) -> str:
        if _move_to is None:
            _move_to = self.write_to
        return os.path.join(_move_to, os.path.split(path)[-1])

    def _set_logger(self) -> None:
        versions = straxen.print_versions(
            modules="strax straxen utilix daqnt numpy tensorflow numba".split(),
            include_git=True,
            return_string=True,
        )
        log_name = "restrax" + self.hostname + ("" if self.production else "_TESTING")
        self.log = daqnt.get_daq_logger(
            log_name,
            log_name,
            level=logging.DEBUG,
            opening_message=f"Welcome to restrax\n{versions}",
        )

    def _move_dir(self, source: str, dest: str) -> None:
        self.log.info(f"Move {source} -> {dest}")
        if self.production:
            os.rename(source, dest)

    def _remove_dir(self, directory: str) -> None:
        """Remove directory (when in production mode)"""
        self.log.info(f"Remove {directory}")
        if self.production:
            shutil.rmtree(directory)

    def log_warning(self, message: str, **kw) -> None:  # type: ignore
        self.log.warning(message)
        for key, value in dict(production=self.production, user=f"restrax_{self.hostname}").items():
            kw.setdefault(key, value)
        super().log_warning(message, **kw)

    @property
    def is_in_bypass_mode(self) -> bool:
        return self.bypass_mode

    def overwrite_settings(self):
        """Overwrite current settings with values from the database."""
        update_config = self.daq_db["restrax_config"].find_one({"name": "restrax_config"})

        type_hints = ty.get_type_hints(self.__class__)

        for field, value in update_config.items():

            if hasattr(self, field):
                current = getattr(self, field)
                if current == value:
                    continue
                current_type = type(current)

                # Match the type hinting from the class to the set value
                matches_type_hints = field in type_hints and isinstance(
                    value, self._get_instance_from_hint(type_hints[field])
                )

                is_same = isinstance(current, type(value))
                both_list = isinstance(current, tuple) and isinstance(value, list)
                both_dict = isinstance(current, immutabledict.immutabledict) and isinstance(
                    value, dict
                )
                if matches_type_hints or is_same or both_list or both_dict:
                    self.log.info(
                        f"Update self.{field} from {current} -> {value}. User"
                        f" {update_config.get('user', '??')}"
                    )
                    setattr(self, field, value)
                else:
                    self.log.warning(
                        f"Not updating {field} to {value} because the type is inconsistent "
                        f"(got {type(value)}, expected {current_type})."
                    )
        self.log.info(f"Done updating defaults (if any)")

    @staticmethod
    def _get_instance_from_hint(hint):
        """Get instances consistent with type hint.

        Searched in vain for a coherent way to typecheck generally in python 3.8.
        Python 3.11 has some nifty features (typing.reveal_type).

        Some examples:

        ```
        # Gives True
        isinstance(dict(), _get_instance_from_hint(ty.Mapping)),
        isinstance([], _get_instance_from_hint(ty.Sequence)),
        isinstance(tuple(), _get_instance_from_hint(ty.Sequence)),
        isinstance(1, _get_instance_from_hint(ty.Union[str, int])),
        isinstance(None, _get_instance_from_hint(ty.Optional[ty.Union[str, int]])),
        isinstance(1, _get_instance_from_hint(ty.Optional[ty.Union[str, int]])),

        # Gives False
        isinstance([], _get_instance_from_hint(ty.Mapping)),
        isinstance(dict(), _get_instance_from_hint(ty.Sequence)),
        isinstance(0.1, _get_instance_from_hint(ty.Sequence)),
        isinstance(0.1, _get_instance_from_hint(ty.Union[str, int])),
        isinstance(dict(), _get_instance_from_hint(ty.Optional[ty.Union[str, int]])),
        ```

        """
        if isinstance(hint, type):
            return hint

        if args := ty.get_args(hint):
            return args

        if origin := ty.get_origin(hint):
            return origin


if __name__ == "__main__":
    main()



from . import peaks_som
from .peaks_som import *

from . import peaks_vanilla
from .peaks_vanilla import *

from . import peak_basics_som
from .peak_basics_som import *

from . import peak_basics_vanilla
from .peak_basics_vanilla import *

from . import peak_positions_cnf
from .peak_positions_cnf import *

from . import peak_positions_mlp
from .peak_positions_mlp import *

from . import peak_positions
from .peak_positions import *

from . import peak_ambience
from .peak_ambience import *

from . import peak_shadow
from .peak_shadow import *

from . import peak_se_score
from .peak_se_score import *

from . import peak_proximity
from .peak_proximity import *

from . import peak_nearest_triggering
from .peak_nearest_triggering import *

from . import peak_top_bottom_params
from .peak_top_bottom_params import *

from . import peak_s1_positions_cnn
from .peak_s1_positions_cnn import *

from . import peak_per_event
from .peak_per_event import *

from . import peak_corrections
from .peak_corrections import *



from . import events
from .events import *

from . import event_basics_som
from .event_basics_som import *

from . import event_basics_vanilla
from .event_basics_vanilla import *

from . import event_positions
from .event_positions import *

from . import event_ambience
from .event_ambience import *

from . import event_shadow
from .event_shadow import *

from . import event_se_score
from .event_se_score import *

from . import veto_proximity
from .veto_proximity import *

from . import distinct_channels
from .distinct_channels import *

from . import energy_estimates
from .energy_estimates import *

from . import corrected_areas
from .corrected_areas import *

from . import event_info
from .event_info import *

from . import event_area_per_channel
from .event_area_per_channel import *

from . import event_pattern_fit
from .event_pattern_fit import *

from . import event_info_double
from .event_info_double import *

from . import event_nearest_triggering
from .event_nearest_triggering import *

from . import event_waveform
from .event_waveform import *

from . import event_top_bottom_params
from .event_top_bottom_params import *

from . import s2_recon_pos_diff
from .s2_recon_pos_diff import *

from . import event_position_uncertainty
from .event_position_uncertainty import *

from . import event_s2_positions_mlp
from .event_s2_positions_mlp import *

from . import event_s1_positions_cnn
from .event_s1_positions_cnn import *

from . import local_minimum_info
from .local_minimum_info import *

from . import multi_scatter
from .multi_scatter import *



import strax
from .peaks_vanilla import PeaksVanilla

export, __all__ = strax.exporter()


@export
class PeakPositionsBase(PeaksVanilla):

    __version__ = "0.0.0"
    child_plugin = True
    save_when = strax.SaveWhen.ALWAYS

    def infer_dtype(self):
        return self.deps[f"peaklet_positions_{self.algorithm}"].dtype_for(
            f"peaklet_positions_{self.algorithm}"
        )

    def compute(self, peaklets, merged_s2s):
        _merged_s2s = strax.merge_arrs([merged_s2s], dtype=peaklets.dtype, replacing=True)
        return super().compute(peaklets, _merged_s2s)



import strax
import straxen
import numpy as np
import numba
from scipy.special import loggamma
from straxen.numbafied_scipy import numba_gammaln, numba_betainc
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class EventPatternFit(strax.Plugin):
    """Plugin that provides patter information for events."""

    depends_on = ("event_area_per_channel", "event_basics", "event_positions")
    provides = "event_pattern_fit"
    __version__ = "0.1.3"

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    # Getting S1 AFT maps
    s1_aft_map = straxen.URLConfig(
        default="itp_map://resource://xedocs://s1_aft_xyz_maps"
        "?attr=value&fmt=json&run_id=plugin.run_id&version=ONLINE",
        cache=True,
    )

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    electron_drift_time_gate = straxen.URLConfig(
        default="xedocs://electron_drift_time_gates?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="Electron drift time from the gate in ns",
        cache=True,
    )

    s1_optical_map = straxen.URLConfig(
        help="S1 (x, y, z) optical/pattern map.",
        infer_type=False,
        default=(
            "itp_map://resource://"
            "XENONnT_s1_xyz_patterns_corrected_qes_MCva43fa9b_wires.pkl"
            "?fmt=pkl"
        ),
    )

    s2_optical_map = straxen.URLConfig(
        help="S2 (x, y) optical/pattern map.",
        infer_type=False,
        default=(
            "itp_map://resource://"
            "XENONnT_s2_xy_patterns_LCE_corrected_qes_MCva43fa9b_wires.pkl"
            "?fmt=pkl"
        ),
    )

    s2_tf_model = straxen.URLConfig(
        help="S2 (x, y) optical data-driven model",
        infer_type=False,
        default=(
            "tf://resource://"
            "XENONnT_s2_optical_map_data_driven_ML_v0_2021_11_25.keras.tar.gz"
            "?custom_objects=plugin.s2_map_custom_objects"
            "&fmt=abs_path&register=True"
        ),
    )

    mean_pe_per_photon = straxen.URLConfig(
        help="Mean of full VUV single photon response",
        default=1.2,
        infer_type=False,
    )

    gain_model = straxen.URLConfig(
        infer_type=False, help="PMT gain model. Specify as (model_type, model_config)"
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    n_top_pmts = straxen.URLConfig(type=int, help="Number of top TPC PMTs")

    s1_min_area_pattern_fit = straxen.URLConfig(
        infer_type=False,
        help="Skip EventPatternFit reconstruction if S1 area (PE) is less than this",
        default=2,
    )

    s2_min_area_pattern_fit = straxen.URLConfig(
        infer_type=False,
        help="Skip EventPatternFit reconstruction if S2 area (PE) is less than this",
        default=10,
    )

    store_per_channel = straxen.URLConfig(
        default=False, type=bool, help="Store normalized LLH per channel for each peak"
    )

    max_r_pattern_fit = straxen.URLConfig(
        default=straxen.tpc_r,
        type=float,
        help="Maximal radius of the peaks where llh calculation will be performed",
    )

    def infer_dtype(self):
        dtype = [
            ("s2_2llh", np.float32, "Modified Poisson likelihood value for main S2 in the event"),
            (
                "s2_neural_2llh",
                np.float32,
                "Data-driven based likelihood value for main S2 in the event",
            ),
            ("alt_s2_2llh", np.float32, "Modified Poisson likelihood value for alternative S2"),
            (
                "alt_s2_neural_2llh",
                np.float32,
                "Data-driven based likelihood value for alternative S2 in the event",
            ),
            ("s1_2llh", np.float32, "Modified Poisson likelihood value for main S1"),
            (
                "s1_top_2llh",
                np.float32,
                "Modified Poisson likelihood value for main S1, calculated from top array",
            ),
            (
                "s1_bottom_2llh",
                np.float32,
                "Modified Poisson likelihood value for main S1, calculated from bottom array",
            ),
            (
                "s1_area_fraction_top_continuous_probability",
                np.float32,
                "Continuous binomial test for S1 area fraction top",
            ),
            (
                "s1_area_fraction_top_discrete_probability",
                np.float32,
                "Discrete binomial test for S1 area fraction top",
            ),
            (
                "s1_photon_fraction_top_continuous_probability",
                np.float32,
                "Continuous binomial test for S1 photon fraction top",
            ),
            (
                "s1_photon_fraction_top_discrete_probability",
                np.float32,
                "Discrete binomial test for S1 photon fraction top",
            ),
            (
                "alt_s1_area_fraction_top_continuous_probability",
                np.float32,
                "Continuous binomial test for alternative S1 area fraction top",
            ),
            (
                "alt_s1_area_fraction_top_discrete_probability",
                np.float32,
                "Discrete binomial test for alternative S1 area fraction top",
            ),
            (
                "alt_s1_photon_fraction_top_continuous_probability",
                np.float32,
                "Continuous binomial test for alternative S1 photon fraction top",
            ),
            (
                "alt_s1_photon_fraction_top_discrete_probability",
                np.float32,
                "Discrete binomial test for alternative S1 photon fraction top",
            ),
        ]

        if self.store_per_channel:
            dtype += [
                (
                    ("2LLH per channel for main S2", "s2_2llh_per_channel"),
                    np.float32,
                    (self.n_top_pmts,),
                ),
                (
                    ("2LLH per channel for alternative S2", "alt_s2_2llh_per_channel"),
                    np.float32,
                    (self.n_top_pmts,),
                ),
                (("Pattern main S2", "s2_pattern"), np.float32, (self.n_top_pmts,)),
                (("Pattern alt S2", "alt_s2_pattern"), np.float32, (self.n_top_pmts,)),
                (("Pattern for main S1", "s1_pattern"), np.float32, (self.n_tpc_pmts,)),
                (
                    ("2LLH per channel for main S1", "s1_2llh_per_channel"),
                    np.float32,
                    (self.n_tpc_pmts,),
                ),
            ]
        dtype += strax.time_fields
        return dtype

    @property
    def s2_map_custom_objects(self):
        def _logl_loss(patterns_true, likelihood):
            return likelihood / 10.0

        return {"_logl_loss": _logl_loss}

    def setup(self):
        # FIXME: Consider renaming the configs to match usage

        self.to_pe = self.gain_model

        self.mean_pe_photon = self.mean_pe_per_photon

        # Getting optical maps
        self.s1_pattern_map = self.s1_optical_map
        self.s2_pattern_map = self.s2_optical_map

        # Getting S2 data-driven tensorflow models
        self.model = self.s2_tf_model

        import tensorflow as tf

        self.model_chi2 = tf.keras.Model(
            self.model.inputs, self.model.get_layer("Likelihood").output
        )

        # Getting gain model to get dead PMTs
        self.dead_PMTs = np.where(self.to_pe == 0)[0]
        self.pmtbool = ~np.in1d(np.arange(0, self.n_tpc_pmts), self.dead_PMTs)
        self.pmtbool_top = self.pmtbool[: self.n_top_pmts]
        self.pmtbool_bottom = self.pmtbool[self.n_top_pmts : self.n_tpc_pmts]

    def compute(self, events):
        result = np.zeros(len(events), dtype=self.dtype)
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)

        # Computing LLH values for S1s
        self.compute_s1_llhvalue(events, result)

        # Computing LLH values for S2s
        self.compute_s2_llhvalue(events, result)

        # Computing chi2 values for S2s
        self.compute_s2_neural_llhvalue(events, result)

        # Computing binomial test for s1 area fraction top
        positions = np.vstack([events["x"], events["y"], events["z"]]).T
        aft_prob = self.s1_aft_map(positions)

        alt_s1_interaction_drift_time = events["s2_center_time"] - events["alt_s1_center_time"]
        alt_s1_interaction_z = -self.electron_drift_velocity * (
            alt_s1_interaction_drift_time - self.electron_drift_time_gate
        )
        alt_positions = np.vstack([events["x"], events["y"], alt_s1_interaction_z]).T
        alt_aft_prob = self.s1_aft_map(alt_positions)

        # main s1 events
        mask_s1 = ~np.isnan(aft_prob)
        mask_s1 &= ~np.isnan(events["s1_area"])
        mask_s1 &= ~np.isnan(events["s1_area_fraction_top"])

        # default value is nan, it will be overwrite if the event satisfy the requirements
        result["s1_area_fraction_top_continuous_probability"][:] = np.nan
        result["s1_area_fraction_top_discrete_probability"][:] = np.nan
        result["s1_photon_fraction_top_continuous_probability"][:] = np.nan
        result["s1_photon_fraction_top_discrete_probability"][:] = np.nan

        # compute binomial test only if we have events that have valid aft prob, s1 area and s1 aft
        if np.sum(mask_s1):
            arg = (
                aft_prob[mask_s1],
                events["s1_area"][mask_s1],
                events["s1_area_fraction_top"][mask_s1],
            )
            result["s1_area_fraction_top_continuous_probability"][mask_s1] = (
                s1_area_fraction_top_probability(*arg)
            )
            result["s1_area_fraction_top_discrete_probability"][mask_s1] = (
                s1_area_fraction_top_probability(*arg, "discrete")
            )
            arg = (
                aft_prob[mask_s1],
                events["s1_area"][mask_s1] / self.mean_pe_per_photon,
                events["s1_area_fraction_top"][mask_s1],
            )
            result["s1_photon_fraction_top_continuous_probability"][mask_s1] = (
                s1_area_fraction_top_probability(*arg)
            )
            result["s1_photon_fraction_top_discrete_probability"][mask_s1] = (
                s1_area_fraction_top_probability(*arg, "discrete")
            )

        # alternative s1 events
        mask_alt_s1 = ~np.isnan(alt_aft_prob)
        mask_alt_s1 &= ~np.isnan(events["alt_s1_area"])
        mask_alt_s1 &= ~np.isnan(events["alt_s1_area_fraction_top"])

        # default value is nan, it will be ovewrite if the event satisfy the requirments
        result["alt_s1_area_fraction_top_continuous_probability"][:] = np.nan
        result["alt_s1_area_fraction_top_discrete_probability"][:] = np.nan
        result["alt_s1_photon_fraction_top_continuous_probability"][:] = np.nan
        result["alt_s1_photon_fraction_top_discrete_probability"][:] = np.nan

        # compute binomial test only if we have events that have valid aft prob,
        # alt s1 area and alt s1 aft
        if np.sum(mask_alt_s1):
            arg = (
                alt_aft_prob[mask_alt_s1],
                events["alt_s1_area"][mask_alt_s1],
                events["alt_s1_area_fraction_top"][mask_alt_s1],
            )
            result["alt_s1_area_fraction_top_continuous_probability"][mask_alt_s1] = (
                s1_area_fraction_top_probability(*arg)
            )
            result["alt_s1_area_fraction_top_discrete_probability"][mask_alt_s1] = (
                s1_area_fraction_top_probability(*arg, "discrete")
            )
            arg = (
                alt_aft_prob[mask_alt_s1],
                events["alt_s1_area"][mask_alt_s1] / self.mean_pe_per_photon,
                events["alt_s1_area_fraction_top"][mask_alt_s1],
            )
            result["alt_s1_photon_fraction_top_continuous_probability"][mask_alt_s1] = (
                s1_area_fraction_top_probability(*arg)
            )
            result["alt_s1_photon_fraction_top_discrete_probability"][mask_alt_s1] = (
                s1_area_fraction_top_probability(*arg, "discrete")
            )

        return result

    def compute_s1_llhvalue(self, events, result):
        # Selecting S1s for pattern fit calculation
        # - must exist (index != -1)
        # - must have total area larger minimal one
        # - must have positive AFT
        x, y, z = events["x"], events["y"], events["z"]
        cur_s1_bool = events["s1_area"] > self.s1_min_area_pattern_fit
        cur_s1_bool &= events["s1_index"] != -1
        cur_s1_bool &= events["s1_area_fraction_top"] >= 0
        cur_s1_bool &= np.isfinite(x)
        cur_s1_bool &= np.isfinite(y)
        cur_s1_bool &= np.isfinite(z)
        cur_s1_bool &= (x**2 + y**2) < self.max_r_pattern_fit**2

        # default value is nan, it will be ovewrite if the event satisfy the requirments
        result["s1_2llh"][:] = np.nan
        result["s1_top_2llh"][:] = np.nan
        result["s1_bottom_2llh"][:] = np.nan

        # Making expectation patterns [ in PE ]
        if np.sum(cur_s1_bool):
            s1_map_effs = self.s1_pattern_map(np.array([x, y, z]).T)[cur_s1_bool, :]
            s1_area = events["s1_area"][cur_s1_bool]
            s1_pattern = (
                s1_area[:, None]
                * (s1_map_effs[:, self.pmtbool])
                / np.sum(s1_map_effs[:, self.pmtbool], axis=1)[:, None]
            )

            s1_pattern_top = events["s1_area_fraction_top"][cur_s1_bool] * s1_area
            s1_pattern_top = s1_pattern_top[:, None] * (
                (s1_map_effs[:, : self.n_top_pmts])[:, self.pmtbool_top]
            )
            s1_pattern_top /= np.sum(
                (s1_map_effs[:, : self.n_top_pmts])[:, self.pmtbool_top], axis=1
            )[:, None]
            s1_pattern_bottom = (1 - events["s1_area_fraction_top"][cur_s1_bool]) * s1_area
            s1_pattern_bottom = s1_pattern_bottom[:, None] * (
                (s1_map_effs[:, self.n_top_pmts :])[:, self.pmtbool_bottom]
            )
            s1_pattern_bottom /= np.sum(
                (s1_map_effs[:, self.n_top_pmts :])[:, self.pmtbool_bottom], axis=1
            )[:, None]

            # Getting pattern from data
            s1_area_per_channel_ = events["s1_area_per_channel"][cur_s1_bool, :]
            s1_area_per_channel = s1_area_per_channel_[:, self.pmtbool]
            s1_area_per_channel_top = (s1_area_per_channel_[:, : self.n_top_pmts])[
                :, self.pmtbool_top
            ]
            s1_area_per_channel_bottom = (s1_area_per_channel_[:, self.n_top_pmts :])[
                :, self.pmtbool_bottom
            ]

            # Top and bottom
            arg1 = s1_pattern / self.mean_pe_photon, s1_area_per_channel, self.mean_pe_photon
            arg2 = (
                s1_area_per_channel / self.mean_pe_photon,
                s1_area_per_channel,
                self.mean_pe_photon,
            )
            norm_llh_val = neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2)
            result["s1_2llh"][cur_s1_bool] = np.sum(norm_llh_val, axis=1)

            # If needed to stire - store only top and bottom array, but not together
            if self.store_per_channel:
                # Storring pattern information
                store_patterns = np.zeros((s1_pattern.shape[0], self.n_tpc_pmts))
                store_patterns[:, self.pmtbool] = s1_pattern
                result["s1_pattern"][cur_s1_bool] = store_patterns
                # Storing actual LLH values
                store_2LLH_ch = np.zeros((norm_llh_val.shape[0], self.n_tpc_pmts))
                store_2LLH_ch[:, self.pmtbool] = norm_llh_val
                result["s1_2llh_per_channel"][cur_s1_bool] = store_2LLH_ch

            # Top
            arg1 = (
                s1_pattern_top / self.mean_pe_photon,
                s1_area_per_channel_top,
                self.mean_pe_photon,
            )
            arg2 = (
                s1_area_per_channel_top / self.mean_pe_photon,
                s1_area_per_channel_top,
                self.mean_pe_photon,
            )
            norm_llh_val = neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2)
            result["s1_top_2llh"][cur_s1_bool] = np.sum(norm_llh_val, axis=1)

            # Bottom
            arg1 = (
                s1_pattern_bottom / self.mean_pe_photon,
                s1_area_per_channel_bottom,
                self.mean_pe_photon,
            )
            arg2 = (
                s1_area_per_channel_bottom / self.mean_pe_photon,
                s1_area_per_channel_bottom,
                self.mean_pe_photon,
            )
            norm_llh_val = neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2)
            result["s1_bottom_2llh"][cur_s1_bool] = np.sum(norm_llh_val, axis=1)

    def compute_s2_llhvalue(self, events, result):
        for t_ in ["s2", "alt_s2"]:
            # Selecting S2s for pattern fit calculation
            # - must exist (index != -1)
            # - must have total area larger minimal one
            # - must have positive AFT
            x, y = events[t_ + "_x"], events[t_ + "_y"]
            s2_mask = events[t_ + "_area"] > self.s2_min_area_pattern_fit
            s2_mask &= events[t_ + "_area_fraction_top"] > 0
            s2_mask &= (x**2 + y**2) < self.max_r_pattern_fit**2

            # default value is nan, it will be ovewrite if the event satisfy the requirments
            result[t_ + "_2llh"][:] = np.nan

            # Making expectation patterns [ in PE ]
            if np.sum(s2_mask):
                s2_map_effs = self.s2_pattern_map(np.array([x, y]).T)[s2_mask, : self.n_top_pmts]
                s2_map_effs = s2_map_effs[:, self.pmtbool_top]
                s2_top_area = (events[t_ + "_area_fraction_top"] * events[t_ + "_area"])[s2_mask]
                s2_pattern = (
                    s2_top_area[:, None] * s2_map_effs / np.sum(s2_map_effs, axis=1)[:, None]
                )

                # Getting pattern from data
                s2_top_area_per_channel = events[t_ + "_area_per_channel"][
                    s2_mask, : self.n_top_pmts
                ]
                s2_top_area_per_channel = s2_top_area_per_channel[:, self.pmtbool_top]

                # Calculating LLH, this is shifted Poisson
                # we get area expectation and we need to scale them to get
                # photon expectation
                norm_llh_val = neg2llh_modpoisson(
                    mu=s2_pattern / self.mean_pe_photon,
                    areas=s2_top_area_per_channel,
                    mean_pe_photon=self.mean_pe_photon,
                ) - neg2llh_modpoisson(
                    mu=s2_top_area_per_channel / self.mean_pe_photon,
                    areas=s2_top_area_per_channel,
                    mean_pe_photon=self.mean_pe_photon,
                )
                result[t_ + "_2llh"][s2_mask] = np.sum(norm_llh_val, axis=1)

                if self.store_per_channel:
                    store_patterns = np.zeros((s2_pattern.shape[0], self.n_top_pmts))
                    store_patterns[:, self.pmtbool_top] = s2_pattern
                    result[t_ + "_pattern"][s2_mask] = store_patterns

                    store_2LLH_ch = np.zeros((norm_llh_val.shape[0], self.n_top_pmts))
                    store_2LLH_ch[:, self.pmtbool_top] = norm_llh_val
                    result[t_ + "_2llh_per_channel"][s2_mask] = store_2LLH_ch

    def compute_s2_neural_llhvalue(self, events, result):
        for t_ in ["s2", "alt_s2"]:
            x, y = events[t_ + "_x"], events[t_ + "_y"]
            s2_mask = events[t_ + "_area"] > self.s2_min_area_pattern_fit
            s2_mask &= events[t_ + "_area_fraction_top"] > 0

            # default value is nan, it will be ovewrite if the event satisfy the requirements
            result[t_ + "_neural_2llh"][:] = np.nan

            # Produce position and top pattern to feed tensorflow model, return chi2/N
            if np.sum(s2_mask):
                s2_pos = np.stack((x, y)).T[s2_mask]
                s2_pat = events[t_ + "_area_per_channel"][s2_mask, : self.n_top_pmts]
                # Output[0]: loss function, -2*log-likelihood, Output[1]: chi2
                result[t_ + "_neural_2llh"][s2_mask] = self.model_chi2.predict(
                    {"xx": s2_pos, "yy": s2_pat}, verbose=0
                )[1]


def neg2llh_modpoisson(mu=None, areas=None, mean_pe_photon=1.0):
    """Modified poisson distribution with proper normalization for shifted poisson.

    mu - expected number of photons per channel
    areas - observed areas per channel
    mean_pe_photon - mean of area responce for one photon

    """
    with np.errstate(divide="ignore", invalid="ignore"):
        fraction = areas / mean_pe_photon
        res = 2.0 * (
            mu - (fraction) * np.log(mu) + loggamma((fraction) + 1) + np.log(mean_pe_photon)
        )
    is_zero = areas <= 0  # If area equals or smaller than 0 - assume 0
    res[is_zero] = 2.0 * mu[is_zero]
    # if zero channel has negative expectation, assume LLH to be 0 there
    # this happens in the normalization factor calculation when mu is received from area
    neg_mu = mu < 0.0
    res[is_zero | neg_mu] = 0.0
    return res


# continuous and discrete binomial test


@numba.njit
def lbinom_pmf(k, n, p):
    """Log of binomial probability mass function approximated with gamma function."""
    scale_log = numba_gammaln(n + 1) - numba_gammaln(n - k + 1) - numba_gammaln(k + 1)
    ret_log = scale_log + k * np.log(p) + (n - k) * np.log(1 - p)
    return ret_log


@numba.njit
def binom_pmf(k, n, p):
    """Binomial probability mass function approximated with gamma function."""
    return np.exp(lbinom_pmf(k, n, p))


@numba.njit
def binom_cdf(k, n, p):
    if k >= n:
        return 1.0
    return numba_betainc(n - k, k + 1, 1.0 - p)


@numba.njit
def binom_sf(k, n, p):
    return 1 - binom_cdf(k, n, p)


@numba.njit
def lbinom_pmf_diriv(k, n, p, dk=1e-7):
    """Numerical dirivitive of Binomial pmf approximated with gamma function."""
    if k + dk < n:
        return (lbinom_pmf(k + dk, n, p) - lbinom_pmf(k, n, p)) / dk
    else:
        return (lbinom_pmf(k - dk, n, p) - lbinom_pmf(k, n, p)) / -dk


@numba.njit(cache=True)
def _numeric_derivative(y0, y1, err, target, x_min, x_max, x0, x1):
    """Get close to <target> by doing a numeric derivative."""
    if abs(y1 - y0) < err:
        # break by passing dx == 0
        if abs(y0 - target) < abs(y1 - target):
            x = x0
        else:
            x = x1
        return 0.0, x, x

    x = (target - y0) / (y1 - y0) * (x1 - x0) + x0
    x = min(x, x_max)
    x = max(x, x_min)

    dx = abs(x - x1)
    x0 = x1
    x1 = x

    return dx, x0, x1


@numba.njit
def lbinom_pmf_mode(x_min, x_max, target, args, err=1e-7, max_iter=50):
    """Find the root of the derivative of log Binomial pmf with secant method."""
    x0 = x_min
    x1 = x_max
    dx = abs(x1 - x0)

    while (dx > err) and (max_iter > 0):
        y0 = lbinom_pmf_diriv(x0, *args)
        y1 = lbinom_pmf_diriv(x1, *args)
        dx, x0, x1 = _numeric_derivative(y0, y1, err, target, x_min, x_max, x0, x1)
        max_iter -= 1
    return x1


@numba.njit
def lbinom_pmf_inverse(x_min, x_max, target, args, err=1e-7, max_iter=50):
    """Find the where the log Binomial pmf cross target with secant method."""
    x0 = x_min
    x1 = x_max
    dx = abs(x1 - x0)

    if dx != 0:
        while (dx > err) and (max_iter > 0):
            y0 = lbinom_pmf(x0, *args)
            y1 = lbinom_pmf(x1, *args)
            dx, x0, x1 = _numeric_derivative(y0, y1, err, target, x_min, x_max, x0, x1)
            max_iter -= 1
        if x0 == x1 == 0 and y0 - target > 0:
            x1 = np.nan
        if x0 == x1 == args[0] and y0 - target < 0:
            x1 = np.nan
    else:
        x1 = np.nan

    return x1


@numba.njit
def binom_test(k, n, p):
    """The main purpose of this algorithm is to find the value j on the other side of the mode that
    has the same probability as k, and integrate the tails outward from k and j.

    In the case where either k or j are zero, only the non-zero tail is integrated.

    """
    mode = lbinom_pmf_mode(0, n, 0, (n, p))
    distance = abs(mode - k)
    target = lbinom_pmf(k, n, p)

    if k < mode:
        j_min = mode
        j_max = min(mode + 2.0 * distance, n)
        j = lbinom_pmf_inverse(j_min, j_max, target, (n, p))
        ls, rs = k, j
    elif k > mode:
        j_min = max(mode - 2.0 * distance, 0)
        j_max = mode
        j = lbinom_pmf_inverse(j_min, j_max, target, (n, p))
        ls, rs = j, k
    else:
        return 1

    pval = 0
    if not np.isnan(ls):
        pval += binom_cdf(ls, n, p)
    if not np.isnan(rs):
        pval += binom_sf(rs, n, p)
        if np.isnan(ls):
            pval += binom_pmf(rs, n, p)

    return pval


@np.vectorize
@numba.njit
def s1_area_fraction_top_probability(aft_prob, area_tot, area_fraction_top, mode="continuous"):
    """Function to compute the S1 AFT probability."""
    area_top = area_tot * area_fraction_top

    # Raise a warning in case one of these three condition is verified
    # and return binomial test equal to nan since they are not physical
    # k: size_top, n: size_tot, p: aft_prob
    do_test = True
    if area_tot < area_top:
        # warnings.warn(f'n {area_tot} must be >= k {area_top}')
        binomial_test = np.nan
        do_test = False
    if (aft_prob > 1.0) or (aft_prob < 0.0):
        # warnings.warn(f'p {aft_prob} must be in range [0, 1]')
        binomial_test = np.nan
        do_test = False
    if area_top < 0:
        # warnings.warn(f'k {area_top} must be >= 0')
        binomial_test = np.nan
        do_test = False

    if do_test:
        if mode == "discrete":
            binomial_test = binom_pmf(area_top, area_tot, aft_prob)
            # TODO:
            # binomial_test = binomtest(
            #     k=round(area_top), n=round(area_tot), p=aft_prob, alternative="two-sided"
            # ).pvalue
        else:
            binomial_test = binom_test(area_top, area_tot, aft_prob)

    return binomial_test



from typing import Optional

import strax
import straxen
import numpy as np
from warnings import warn

export, __all__ = strax.exporter()


@export
class EventS2PositionBase(strax.Plugin):
    """Base pluging for S2 position reconstruction at event level."""

    __version__ = "0.0.0"
    depends_on = ("event_area_per_channel", "event_basics")

    algorithm: Optional[str] = None
    compressor = "zstd"

    min_reconstruction_area = straxen.URLConfig(
        help="Skip reconstruction if area (PE) is less than this",
        default=0,
        infer_type=False,
    )
    n_top_pmts = straxen.URLConfig(
        default=straxen.n_top_pmts, infer_type=False, help="Number of top PMTs"
    )

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(
                f"Base class should not be used without algorithm as done in {__class__.__name__}"
            )

        dtype = [
            (
                "event_s2_x_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S2 X position (cm), uncorrected",
            ),
            (
                "event_s2_y_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S2 Y position (cm), uncorrected",
            ),
            (
                "event_alt_s2_x_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} alt S2 X position (cm), uncorrected",
            ),
            (
                "event_alt_s2_y_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} alt S2 Y position (cm), uncorrected",
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """Simple wrapper to have several tf_model_mlp, tf_model_cnf, ..

        point to this same function in the compute method

        """
        model = getattr(self, f"tf_model_{self.algorithm}", None)
        if model is None:
            warn(
                f"Setting model to None for {self.__class__.__name__} will "
                f"set only nans as output for {self.algorithm}"
            )
        if isinstance(model, str):
            raise ValueError(
                f"open files from tf:// protocol! Got {model} "
                "instead, see tests/test_posrec.py for examples."
            )
        return model

    def compute(self, events):
        result = np.ones(len(events), dtype=self.dtype)
        result["time"], result["endtime"] = events["time"], strax.endtime(events)
        result["event_s2_x_" + self.algorithm] *= np.nan
        result["event_s2_y_" + self.algorithm] *= np.nan
        result["event_alt_s2_x_" + self.algorithm] *= np.nan
        result["event_alt_s2_y_" + self.algorithm] *= np.nan

        model = self.get_tf_model()

        for p_type in ["s2", "alt_s2"]:
            peak_mask = events[p_type + "_area"] > self.min_reconstruction_area
            if not np.sum(peak_mask):
                continue

            _top_pattern = events[p_type + "_area_per_channel"][peak_mask, : self.n_top_pmts]
            with np.errstate(divide="ignore", invalid="ignore"):
                _top_pattern = _top_pattern / np.max(_top_pattern, axis=1).reshape(-1, 1)

            _top_pattern = _top_pattern.reshape(-1, self.n_top_pmts)
            _pos = model.predict(_top_pattern)

            result["event_" + p_type + "_x_" + self.algorithm][peak_mask] = _pos[:, 0]
            result["event_" + p_type + "_y_" + self.algorithm][peak_mask] = _pos[:, 1]

        return result



import strax

from straxen.plugins.peaklets.peaklet_classification_vanilla import PeakletClassificationVanilla
from straxen.plugins.defaults import HE_PREAMBLE

export, __all__ = strax.exporter()


@export
class PeakletClassificationHighEnergy(PeakletClassificationVanilla):
    __doc__ = HE_PREAMBLE + (PeakletClassificationVanilla.__doc__ or "")

    __version__ = "0.0.2"
    child_plugin = True
    depends_on = "peaklets_he"
    provides = "peaklet_classification_he"

    def compute(self, peaklets_he):
        return super().compute(peaklets_he)



from . import peaklets
from .peaklets import *

from . import peaklet_classification_som
from .peaklet_classification_som import *

from . import peaklet_classification_vanilla
from .peaklet_classification_vanilla import *

from . import peaklet_positions_cnf
from .peaklet_positions_cnf import *

from . import peaklet_positions_mlp
from .peaklet_positions_mlp import *
